import pytest

from gobbli.dataset.cmu_movie_summary import MovieSummaryDataset
from gobbli.dataset.newsgroups import NewsgroupsDataset
from gobbli.dataset.trivial import TrivialDataset
from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText
from gobbli.model.majority import MajorityClassifier
from gobbli.model.mtdnn import MTDNN
from gobbli.model.sklearn import SKLearnClassifier
from gobbli.model.spacy import SpaCyModel
from gobbli.model.transformer import Transformer
from gobbli.test.util import model_test_dir, skip_if_low_resource, validate_checkpoint


def check_predict_output(train_output, predict_input, predict_output):
    assert predict_output.y_pred_proba.shape == (
        len(predict_input.X),
        len(train_output.labels),
    )

    assert len(predict_output.y_pred) == len(predict_input.X)
    label_set = set(train_output.labels)
    for p in predict_output.y_pred:
        assert p in label_set


@pytest.mark.parametrize(
    "model_cls,dataset_cls,model_kwargs,train_kwargs,predict_kwargs",
    [
        (MajorityClassifier, TrivialDataset, {}, {}, {}),
        (MajorityClassifier, NewsgroupsDataset, {}, {}, {}),
        (MajorityClassifier, MovieSummaryDataset, {}, {}, {}),
        (SKLearnClassifier, TrivialDataset, {}, {}, {}),
        (SKLearnClassifier, NewsgroupsDataset, {}, {}, {}),
        (SKLearnClassifier, MovieSummaryDataset, {}, {}, {}),
        (
            BERT,
            TrivialDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"predict_batch_size": 1},
        ),
        (
            BERT,
            NewsgroupsDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 8},
            {"predict_batch_size": 32},
        ),
        (
            BERT,
            MovieSummaryDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 8},
            {"predict_batch_size": 32},
        ),
        (
            MTDNN,
            TrivialDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"predict_batch_size": 1},
        ),
        (
            MTDNN,
            NewsgroupsDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            MTDNN,
            MovieSummaryDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            FastText,
            TrivialDataset,
            {"autotune_duration": 10, "word_ngrams": 1, "dim": 50, "ws": 5},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"predict_batch_size": 1},
        ),
        (
            FastText,
            NewsgroupsDataset,
            {"autotune_duration": 10, "word_ngrams": 1, "dim": 50, "ws": 5},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            FastText,
            MovieSummaryDataset,
            {"autotune_duration": 10, "word_ngrams": 1, "dim": 50, "ws": 5},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            Transformer,
            TrivialDataset,
            {"max_seq_length": 128},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"predict_batch_size": 1},
        ),
        (
            Transformer,
            NewsgroupsDataset,
            {"max_seq_length": 128},
            {"num_train_epochs": 1, "train_batch_size": 16, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            Transformer,
            MovieSummaryDataset,
            {"max_seq_length": 128},
            {"num_train_epochs": 1, "train_batch_size": 16, "valid_batch_size": 32},
            {"predict_batch_size": 32},
        ),
        (
            SpaCyModel,
            TrivialDataset,
            {"model": "en_core_web_sm", "architecture": "bow"},
            {"num_train_epochs": 1, "train_batch_size": 1},
            {},
        ),
        (
            SpaCyModel,
            NewsgroupsDataset,
            {"model": "en_core_web_sm", "architecture": "bow"},
            {"num_train_epochs": 1, "train_batch_size": 32},
            {},
        ),
        (
            SpaCyModel,
            MovieSummaryDataset,
            {"model": "en_core_web_sm", "architecture": "bow"},
            {"num_train_epochs": 1, "train_batch_size": 32},
            {},
        ),
    ],
)
def test_classifier(
    model_cls,
    dataset_cls,
    model_kwargs,
    train_kwargs,
    predict_kwargs,
    model_gpu_config,
    gobbli_dir,
    request,
):
    """
    Ensure classifiers train and predict appropriately across a few example datasets.
    """
    # These combinations of model and dataset require a lot of memory
    if model_cls in (BERT, MTDNN, Transformer) and dataset_cls in (
        NewsgroupsDataset,
        MovieSummaryDataset,
    ):
        skip_if_low_resource(request.config)

    model = model_cls(
        data_dir=model_test_dir(model_cls),
        load_existing=True,
        **model_gpu_config,
        **model_kwargs,
    )

    model.build()
    ds = dataset_cls.load()

    train_input = ds.train_input(limit=50, **train_kwargs)
    if train_input.multilabel and model_cls in (BERT, MTDNN):
        pytest.xfail(
            f"model {model_cls.__name__} doesn't support multilabel classification"
        )

    # Verify training runs, results are sensible
    train_output = model.train(train_input)
    assert train_output.valid_loss is not None
    assert train_output.train_loss is not None
    assert 0 <= train_output.valid_accuracy <= 1

    validate_checkpoint(model.__class__, train_output.checkpoint)

    predict_input = ds.predict_input(limit=50, **predict_kwargs)

    if isinstance(model, FastText):
        # fastText requires a trained checkpoint for prediction
        pass
    else:
        # Verify prediction runs without a trained checkpoint
        predict_output = model.predict(predict_input)
        check_predict_output(train_output, predict_input, predict_output)

    # Verify prediction runs with a trained checkpoint
    predict_input.checkpoint = train_output.checkpoint
    predict_output = model.predict(predict_input)
    check_predict_output(train_output, predict_input, predict_output)



from gobbli.augment.bert import BERTMaskedLM
from gobbli.augment.marian import MarianMT
from gobbli.augment.word2vec import Word2Vec
from gobbli.augment.wordnet import WordNet

__all__ = ["BERTMaskedLM", "Word2Vec", "WordNet", "MarianMT"]



import json
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

import gobbli.io
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, PredictMixin, TrainMixin
from gobbli.util import assert_type, escape_line_delimited_texts


class SpaCyModel(BaseModel, TrainMixin, PredictMixin, EmbedMixin):
    """
    gobbli interface for spaCy language models which allows for training
    and prediction via the
    `TextCategorizer pipeline component <https://spacy.io/api/textcategorizer>`__
    and static embeddings via `Vectors <https://spacy.io/api/vectors>`__.
    """

    _BUILD_PATH = Path(__file__).parent

    _TRAIN_INPUT_FILE = "train.tsv"
    _VALID_INPUT_FILE = "dev.tsv"
    _TEST_INPUT_FILE = "test.tsv"
    _LABELS_INPUT_FILE = "labels.tsv"

    _TRAIN_OUTPUT_CHECKPOINT = "checkpoint"
    _VALID_OUTPUT_FILE = "valid_results.json"
    _TEST_OUTPUT_FILE = "test_results.tsv"

    _EMBEDDING_INPUT_FILE = "input.tsv"
    _EMBEDDING_OUTPUT_FILE = "embeddings.jsonl"

    _CONTAINER_CACHE_DIR = Path("/cache")

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        spaCy parameters:

        - ``model`` (:obj:`str`): Name of a spaCy model to use.
          Available values are in `the spaCy model docs <https://spacy.io/models>`__ and
          `the spacy-transformers docs <https://github.com/explosion/spacy-transformers>`__.
        - ``architecture`` (:obj:`str`): Model architecture to use.
          Available values are in `the spaCy API docs <https://spacy.io/api/textcategorizer#architectures>`__.
          This is ignored if using a spacy-transformers model.
        - ``dropout`` (:obj:`float`): Dropout proportion for training.
        - ``full_pipeline`` (:obj:`bool`): If True, enable the full spaCy language pipeline
          (including tagging, parsing, and named entity recognition) for the TextCategorizer
          model used in training and prediction.  This makes training/prediction much slower
          but theoretically provides more information to the model.  This is ignored if using a
          spacy-transformers model.

        Note that gobbli relies on spaCy to perform validation on these parameters,
        so initialization errors may not be caught until model runtime.
        """
        self.model = "en_core_web_lg"
        self.architecture = "ensemble"
        self.dropout = 0.2
        self.full_pipeline = False

        for name, value in params.items():
            if name == "model":
                self.model = value
            elif name == "architecture":
                self.architecture = value
            elif name == "dropout":
                assert_type(name, value, float)
                self.dropout = value
            elif name == "full_pipeline":
                assert_type(name, value, bool)
                self.full_pipeline = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the spaCy container.
        """
        return "gobbli-spacy"

    def _build(self):
        # Add the spaCy model to the image build so it's properly installed
        base_build_kwargs = deepcopy(self._base_docker_build_kwargs)
        if "buildargs" not in base_build_kwargs:
            base_build_kwargs["buildargs"] = {}
        base_build_kwargs["buildargs"]["model"] = self.model

        self.docker_client.images.build(
            path=str(SpaCyModel._BUILD_PATH), tag=self.image_tag, **base_build_kwargs
        )

    @staticmethod
    def _get_checkpoint(
        user_checkpoint: Optional[Path], context: ContainerTaskContext
    ) -> Tuple[Optional[Path], Optional[Path]]:
        """
        Determines the host checkpoint directory and container checkpoint directory
        using the user-requested checkpoint (if any) and the container context.

        Args:
          user_checkpoint: An optional checkpoint passed in by the user.  If the user doesn't
            pass one, use the default pretrained checkpoint.
          context: The container context to create the checkpoint in.

        Returns:
          A 2-tuple: the host checkpoint directory (if any) and
            the container checkpoint directory (if any)
        """
        if user_checkpoint is None:
            host_checkpoint_dir = None
            container_checkpoint_dir = None
        else:
            host_checkpoint_dir = user_checkpoint
            container_checkpoint_dir = context.container_root_dir / "checkpoint"

        return host_checkpoint_dir, container_checkpoint_dir

    def _get_model(self, container_checkpoint_dir: Optional[Path]) -> Union[str, Path]:
        """
        Determine the model to pass to the run_spacy script.  If we don't have a
        checkpoint, we'll use our stock model.  Otherwise, we should use the
        checkpoint.
        """
        if container_checkpoint_dir is None:
            return self.model
        else:
            return container_checkpoint_dir

    @property
    def host_cache_dir(self):
        """
        Directory to be used for downloaded spaCy files.
        Should be the same across all instances of the class, since these are
        generally static model weights that can be reused.
        """
        cache_dir = SpaCyModel.model_class_dir() / "cache"
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    def _write_input(
        self, X: List[str], labels: Optional[List[List[str]]], input_path: Path
    ):
        """
        Write the given input texts and (optionally) labels to the file pointed to by
        ``input_path``.
        """
        df = pd.DataFrame({"Text": X})

        if labels is not None:
            df["Label"] = labels

        df.to_csv(input_path, sep="\t", index=False)

    def _write_labels(self, labels: List[str], labels_path: Path):
        """
        Write the given labels to the file pointed at by ``labels_path``.
        """
        labels_path.write_text(escape_line_delimited_texts(labels))

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:

        if train_input.valid_batch_size != gobbli.io.TrainInput.valid_batch_size:
            warnings.warn(
                "The spaCy model doesn't batch validation data, so the validation "
                "batch size parameter will be ignored."
            )

        self._write_input(
            train_input.X_train,
            train_input.y_train_multilabel,
            context.host_input_dir / SpaCyModel._TRAIN_INPUT_FILE,
        )
        self._write_input(
            train_input.X_valid,
            train_input.y_valid_multilabel,
            context.host_input_dir / SpaCyModel._VALID_INPUT_FILE,
        )

        labels = train_input.labels()
        self._write_labels(
            labels, context.host_input_dir / SpaCyModel._LABELS_INPUT_FILE
        )

        # Determine checkpoint to use
        host_checkpoint_dir, container_checkpoint_dir = self._get_checkpoint(
            train_input.checkpoint, context
        )

        cmd = (
            "python3 run_spacy.py"
            " train"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --model {self._get_model(container_checkpoint_dir)}"
            f" --architecture {self.architecture}"
            f" --cache-dir {SpaCyModel._CONTAINER_CACHE_DIR}"
            f" --train-batch-size {train_input.train_batch_size}"
            f" --num-train-epochs {train_input.num_train_epochs}"
            f" --dropout {self.dropout}"
        )

        if self.full_pipeline:
            cmd += " --full-pipeline"
        if train_input.multilabel:
            cmd += " --multilabel"

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, SpaCyModel._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Read in the generated evaluation results
        with open(context.host_output_dir / SpaCyModel._VALID_OUTPUT_FILE, "r") as f:
            results = json.load(f)

        return gobbli.io.TrainOutput(
            valid_loss=results["mean_valid_loss"],
            valid_accuracy=results["valid_accuracy"],
            train_loss=results["mean_train_loss"],
            labels=labels,
            multilabel=train_input.multilabel,
            checkpoint=context.host_output_dir / SpaCyModel._TRAIN_OUTPUT_CHECKPOINT,
            _console_output=container_logs,
        )

    def _read_predictions(self, predict_path: Path):
        return pd.read_csv(predict_path, sep="\t")

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:

        if (
            predict_input.predict_batch_size
            != gobbli.io.PredictInput.predict_batch_size
        ):
            warnings.warn(
                "The spaCy model doesn't batch prediction data, so the prediction "
                "batch size parameter will be ignored."
            )

        self._write_input(
            predict_input.X, None, context.host_input_dir / SpaCyModel._TEST_INPUT_FILE
        )

        labels = predict_input.labels
        self._write_labels(
            labels, context.host_input_dir / SpaCyModel._LABELS_INPUT_FILE
        )

        host_checkpoint_dir, container_checkpoint_dir = self._get_checkpoint(
            predict_input.checkpoint, context
        )

        cmd = (
            "python3 run_spacy.py"
            " predict"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --model {self._get_model(container_checkpoint_dir)}"
            f" --architecture {self.architecture}"
            f" --cache-dir {SpaCyModel._CONTAINER_CACHE_DIR}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_dir, container_checkpoint_dir
        )

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, SpaCyModel._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        return gobbli.io.PredictOutput(
            y_pred_proba=self._read_predictions(
                context.host_output_dir / SpaCyModel._TEST_OUTPUT_FILE
            ),
            _console_output=container_logs,
        )

    def _read_embeddings(
        self, embed_path: Path, pooling: gobbli.io.EmbedPooling
    ) -> Tuple[List[np.ndarray], Optional[List[List[str]]]]:

        embeddings = []  # type: List[np.ndarray]
        doc_tokens = []  # type: List[List[str]]
        with open(embed_path, "r") as f:
            for line in f:
                line_json = json.loads(line)
                embeddings.append(np.array(line_json["embedding"]))
                if pooling == gobbli.io.EmbedPooling.NONE:
                    doc_tokens.append(line_json["tokens"])

        tokens = None
        if pooling == gobbli.io.EmbedPooling.NONE:
            tokens = doc_tokens

        return embeddings, tokens

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        self._write_input(
            embed_input.X,
            None,
            context.host_input_dir / SpaCyModel._EMBEDDING_INPUT_FILE,
        )

        if embed_input.checkpoint is not None:
            warnings.warn(
                "The spaCy model vectors can't be fine-tuned, so custom "
                "checkpoints are ignored when generating embeddings."
            )

        cmd = (
            "python3 run_spacy.py"
            " embed"
            f" --input-dir {context.container_input_dir}"
            f" --output-dir {context.container_output_dir}"
            f" --model {self.model}"
            f" --architecture {self.architecture}"
            f" --cache-dir {SpaCyModel._CONTAINER_CACHE_DIR}"
            f" --embed-pooling {embed_input.pooling.value}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the cache directory
        maybe_mount(
            run_kwargs["volumes"], self.host_cache_dir, SpaCyModel._CONTAINER_CACHE_DIR
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        X_embedded, embed_tokens = self._read_embeddings(
            context.host_output_dir / SpaCyModel._EMBEDDING_OUTPUT_FILE,
            embed_input.pooling,
        )

        return gobbli.io.EmbedOutput(
            X_embedded=X_embedded,
            embed_tokens=embed_tokens,
            _console_output=container_logs,
        )



import re
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional

import altair as alt
import click
import pandas as pd
import streamlit as st

from gobbli.inspect.evaluate import ClassificationError, ClassificationEvaluation
from gobbli.interactive.util import (
    DEFAULT_PREDICT_BATCH_SIZE,
    get_label_indices,
    get_predictions,
    load_data,
    safe_sample,
    st_model_metadata,
    st_sample_data,
    st_select_model_checkpoint,
)
from gobbli.util import truncate_text


def show_metrics(metrics: Dict[str, Any]):
    st.header("Metrics")
    md = ""
    for name, value in metrics.items():
        md += f"- **{name}:** {value:.4f}\n"
    st.markdown(md)


def show_plot(plot: alt.Chart):
    st.header("Model Predicted Probability by True Class")
    st.altair_chart(plot)


TRUE_LABEL_COLOR = "#1f78b4"
TRUE_LABEL_TEXT_COLOR = "white"

PRED_PROB_LABEL_RE = re.compile(r"^(.+) \((?:[0-9.]+)\)$")


def _show_example_predictions(
    texts: List[str],
    labels: Optional[List[str]],
    y_pred_proba: pd.DataFrame,
    truncate_len: int,
    top_k: int,
):
    def gather_predictions(row):
        ndx = row.name
        pred_prob_order = row.sort_values(ascending=False)[:top_k]
        data = {"Document": truncate_text(texts[ndx], truncate_len)}
        if labels is not None:
            y_true = labels[ndx]
            column_header = "True Labels" if isinstance(y_true, list) else "True Label"
            data[column_header] = y_true

        for i, (label, pred_prob) in enumerate(pred_prob_order.items()):
            data[f"Predicted Label {i+1}"] = f"{label} ({pred_prob:.3f})"

        return pd.Series(data)

    df = y_pred_proba.apply(gather_predictions, axis=1)

    def style_pred_prob(row, labels):
        ndx = row.name
        true_label_style = (
            f"background-color: {TRUE_LABEL_COLOR};color: {TRUE_LABEL_TEXT_COLOR}"
        )
        style = [
            # Text
            "",
            # True label
            true_label_style,
        ]

        pred_probs = row[2:]
        for p in pred_probs:
            match = re.match(PRED_PROB_LABEL_RE, p)
            if match is None:
                raise ValueError(f"Failed to parse predicted probability cell: {p}")

            y_true = labels[ndx]
            if isinstance(y_true, list):
                is_match = match.group(1) in y_true
            else:
                is_match = match.group(1) == y_true

            if is_match:
                # The cell corresponds to a correct label
                cell_style = true_label_style
            else:
                cell_style = ""
            style.append(cell_style)

        return style

    if labels is not None:
        df = df.style.apply(style_pred_prob, axis=1, labels=labels)

    st.table(df)


def show_example_predictions(
    texts: List[str],
    labels: Optional[List[str]],
    y_pred_proba: pd.DataFrame,
    example_truncate_len: int,
    example_num_docs: int,
    example_top_k: int,
):
    st.header("Example Predictions")

    example_indices = safe_sample(range(len(texts)), example_num_docs)
    _show_example_predictions(
        [texts[i] for i in example_indices],
        None if labels is None else [labels[i] for i in example_indices],
        y_pred_proba.iloc[example_indices, :].reset_index(drop=True),
        example_truncate_len,
        example_top_k,
    )


def show_errors(errors: List[ClassificationError], truncate_len: int = 500):
    df_data = []
    for e in errors:
        pred_class = max(e.y_pred_proba, key=e.y_pred_proba.get)
        pred_class_prob = e.y_pred_proba[pred_class]
        true_column_header = (
            "True Labels" if isinstance(e.y_true, list) else "True Label"
        )
        df_data.append(
            # Use OrderedDict to preserve column order
            OrderedDict(
                {
                    "Document": truncate_text(e.X, truncate_len),
                    true_column_header: e.y_true,
                    "Top Predicted Label": f"{pred_class} ({pred_class_prob:.4f})",
                }
            )
        )
    df = pd.DataFrame(df_data)
    st.table(df)


@click.command()
@click.argument("model_data_dir", type=str)
@click.argument("data", type=str)
@click.option(
    "--n-rows",
    type=int,
    help="Number of rows to load from the data file.  If -1, load all rows.",
    default=-1,
    show_default=True,
)
@click.option(
    "--use-gpu/--use-cpu",
    default=False,
    help="Which device to run the model on. Defaults to CPU.",
)
@click.option(
    "--multilabel/--multiclass",
    default=False,
    help="Only applies when reading user data from a file.  If --multilabel, each "
    "observation may have multiple associated labels.  If --multiclass, only one label "
    "is allowed per observation.  Defaults to --multiclass.",
)
@click.option(
    "--multilabel-sep",
    default=",",
    help="Determines how the labels in the label column are separated for a multilabel "
    "dataset read from a file.",
    show_default=True,
)
@click.option(
    "--nvidia-visible-devices",
    default="all",
    help="Which GPUs to make available to the container; ignored if running on CPU. "
    "If not 'all', should be a comma-separated string: ex. ``1,2``.",
    show_default=True,
)
def run(
    model_data_dir: str,
    data: str,
    n_rows: int,
    use_gpu: bool,
    multilabel: bool,
    multilabel_sep: str,
    nvidia_visible_devices: str,
):
    st.sidebar.header("Model")
    model_data_path = Path(model_data_dir)
    model_cls, model_kwargs, checkpoint_meta = st_select_model_checkpoint(
        model_data_path, use_gpu, nvidia_visible_devices
    )
    st.title(f"Evaluating: {model_cls.__name__} on {data}")

    model = model_cls(**model_kwargs)
    st_model_metadata(model)

    texts, labels = load_data(
        data,
        multilabel,
        n_rows=None if n_rows == -1 else n_rows,
        multilabel_sep=multilabel_sep,
    )
    if labels is not None:
        label_indices = get_label_indices(labels)

    checkpoint_labels = checkpoint_meta["labels"]
    num_labels = len(checkpoint_labels)

    if labels is not None:
        dataset_labels = set(label_indices.keys())
        if dataset_labels > set(checkpoint_labels):
            st.warning(
                "The dataset contains some labels that the model was not trained on.  No "
                "predictions will be made for these labels: "
                f"{dataset_labels - set(checkpoint_labels)}"
            )

    sampled_texts, sampled_labels = st_sample_data(texts, labels)

    st.sidebar.header("Predictions")
    predict_batch_size = st.sidebar.number_input(
        "Batch Size",
        min_value=1,
        max_value=len(sampled_texts),
        value=min(len(sampled_texts), DEFAULT_PREDICT_BATCH_SIZE),
    )

    y_pred_proba = get_predictions(
        model_cls,
        model_kwargs,
        sampled_texts,
        checkpoint_labels,
        checkpoint_meta["checkpoint"],
        batch_size=predict_batch_size,
    )

    evaluation = None

    if labels is not None:
        evaluation = ClassificationEvaluation(
            labels=checkpoint_labels,
            X=sampled_texts,
            y_true=sampled_labels,
            y_pred_proba=y_pred_proba,
        )

        show_metrics(evaluation.metrics())
        show_plot(evaluation.plot())

    st.sidebar.header("Examples")
    example_top_k = st.sidebar.number_input(
        "Top K Predictions to Show",
        min_value=1,
        max_value=num_labels,
        value=min(num_labels, 3),
    )
    example_num_docs = st.sidebar.number_input(
        "Number of Example Documents to Show", min_value=1, max_value=None, value=5
    )
    example_truncate_len = st.sidebar.number_input(
        "Example Document Truncate Length", min_value=1, max_value=None, value=500
    )

    show_example_predictions(
        sampled_texts,
        sampled_labels,
        y_pred_proba,
        example_truncate_len,
        example_num_docs,
        example_top_k,
    )

    if labels is not None and evaluation is not None:
        st.sidebar.header("Errors")
        errors_label = st.sidebar.selectbox(
            "Label to Show Errors For", options=checkpoint_labels
        )
        errors_num_docs = st.sidebar.number_input(
            "Number of Error Documents to Show", min_value=1, max_value=None, value=5
        )
        errors_truncate_len = st.sidebar.number_input(
            "Error Document Truncate Length", min_value=1, max_value=None, value=500
        )

        false_positives, false_negatives = evaluation.errors_for_label(
            errors_label, k=errors_num_docs
        )

        st.header(f"Top Model Errors: {errors_label}")

        st.subheader("False Positives")
        show_errors(false_positives, truncate_len=errors_truncate_len)

        st.subheader("False Negatives")
        show_errors(false_negatives, truncate_len=errors_truncate_len)

    if labels is None:
        st.warning(
            "No ground truth labels found for the passed dataset.  Evaluation "
            "metrics and errors can't be calculated without ground truth labels."
        )


if __name__ == "__main__":
    # Streamlit doesn't like click exiting the script early after the function runs
    try:
        run()
    except SystemExit:
        pass



# Copyright (c) Microsoft. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy
from torch.nn.utils import weight_norm
from torch.nn.parameter import Parameter
from .common import activation
from .dropout_wrapper import DropoutWrapper

class DotProduct(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='sim', opt={}, dropout=None):
        super(DotProduct, self).__init__()
        assert x1_dim == x2_dim
        self.opt = opt
        self.prefix = prefix
        self.scale_on = opt.get('{}_scale'.format(self.prefix), False)
        self.scalor = 1.0 / numpy.power(x2_dim, 0.5)

    def forward(self, x1, x2):
        assert x1.size(2) == x2.size(2)
        scores = x1.bmm(x2.transpose(1, 2))
        if self.scale_on:
            scores *= self.scalor
        return scores


class DotProductProject(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='sim', opt={}, dropout=None):
        super(DotProductProject, self).__init__()
        self.prefix = prefix
        self.opt = opt
        self.hidden_size = opt.get('{}_hidden_size'.format(self.prefix), 64)
        self.residual_on = opt.get('{}_residual_on'.format(self.prefix), False)
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.share = opt.get('{}_share'.format(self.prefix), False)
        self.f = activation(opt.get('{}_activation'.format(self.prefix), 'relu'))
        self.scale_on = opt.get('{}_scale_on'.format(self.prefix), False)
        self.dropout = dropout
        x1_in_dim = x1_dim
        x2_in_dim = x2_dim
        out_dim = self.hidden_size
        self.proj_1 = nn.Linear(x1_in_dim, out_dim, bias=False)
        if self.layer_norm_on:
            self.proj_1 = weight_norm(self.proj_1)
        if self.share and x1_in_dim == x2_in_dim:
            self.proj_2 = self.proj_1
        else:
            self.proj_2 = nn.Linear(x2_in_dim, out_dim)
            if self.layer_norm_on:
                self.proj_2 = weight_norm(self.proj_2)

        if self.scale_on:
            self.scalar = Parameter(torch.ones(1,1,1) / (self.hidden_size ** 0.5), requires_grad=False)
        else:
            self.sclalar = Parameter(torch.ones(1,1, self.hidden_size), requires_grad=True)

    def forward(self, x1, x2):
        assert x1.size(2) == x2.size(2)
        if self.dropout:
            x1 = self.dropout(x1)
            x2 = self.dropout(x2)
        x1_flat = x1.contiguous().view(-1, x1.size(2))
        x2_flat = x2.contiguous().view(-1, x2.size(2))
        x1_o = self.f(self.proj_1(x1_flat)).view(x1.size(0), x1.size(1), -1)
        # x2_o = self.f(self.proj_1(x2_flat)).view(x2.size(0), x2.size(1), -1)
        x2_o = self.f(self.proj_2(x2_flat)).view(x2.size(0), x2.size(1), -1)
        if self.scale_on:
            scalar = self.scalar.expand_as(x2_o)
            x2_o = scalar * x2_o
        scores = x1_o.bmm(x2_o.transpose(1, 2))
        return scores


class Bilinear(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='sim', opt={}, dropout=None):
        super(Bilinear, self).__init__()
        self.opt = opt
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.transform_on = opt.get('{}_proj_on'.format(self.prefix), False)
        # self.init = init_wrapper(opt.get('{}_init'.format(self.prefix), ''))
        self.dropout = dropout
        if self.transform_on:
            self.proj = nn.Linear(x1_dim, x2_dim)
            # self.init(self.proj.weight)
            if self.layer_norm_on: self.proj = weight_norm(self.proj)

    def forward(self, x, y):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        if self.dropout:
            x = self.dropout(x)
            y = self.dropout(y)

        proj = self.proj(y) if self.transform_on else y
        if self.dropout:
            proj = self.dropout(proj)
        scores = x.bmm(proj.unsqueeze(2)).squeeze(2)
        return scores


class BilinearSum(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='sim', opt={}, dropout=None):
        super(BilinearSum, self).__init__()
        self.x_linear = nn.Linear(x1_dim, 1, bias=False)
        self.y_linear = nn.Linear(x2_dim, 1, bias=False)
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.init = init_wrapper(opt.get('{}_init'.format(self.prefix), False))
        if self.layer_norm_on:
            self.x_linear = weight_norm(self.x_linear)
            self.y_linear = weight_norm(self.y_linear)

        self.init(self.x_linear.weight)
        self.init(self.y_linear.weight)
        self.dropout = dropout

    def forward(self, x1, x2):
        """
        x1: batch * len1 * input_size
        x2: batch * len2 * input_size
        score: batch * len1 * len2
        """
        if self.dropout:
            x1 = self.dropout(x1)
            x2 = self.dropout(x2)

        x1_logits = self.x_linear(x1.contiguous().view(-1, x1.size(-1))).view(x1.size(0), -1, 1)
        x2_logits = self.y_linear(x2.contiguous().view(-1, x2.size(-1))).view(x2.size(0), 1, -1)

        shape = (x1.size(0), x1.size(1), x2.size())
        scores = x1_logits.expand_as(shape) + x2_logits.expand_as(shape)
        return scores


class Trilinear(nn.Module):
    """Function used in BiDAF"""
    def __init__(self, x1_dim, x2_dim, prefix='sim', opt={}, dropout=None):
        super(Trilinear, self).__init__()
        self.prefix = prefix
        self.x_linear = nn.Linear(x1_dim, 1, bias=False)
        self.x_dot_linear = nn.Linear(x1_dim, 1, bias=False)
        self.y_linear = nn.Linear(x2_dim, 1, bias=False)
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.init = init_wrapper(opt.get('{}_init'.format(self.prefix), 'xavier_uniform'))
        if self.layer_norm_on:
            self.x_linear = weight_norm(self.x_linear)
            self.x_dot_linear = weight_norm(self.x_dot_linear)
            self.y_linear = weight_norm(self.y_linear)

        self.init(self.x_linear.weight)
        self.init(self.x_dot_linear.weight)
        self.init(self.y_linear.weight)
        self.dropout = dropout

    def forward(self, x1, x2):
        """
        x1: batch * len1 * input_size
        x2: batch * len2 * input_size
        score: batch * len1 * len2
        """
        if self.dropout:
            x1 = self.dropout(x1)
            x2 = self.dropout(x2)

        x1_logits = self.x_linear(x1.contiguous().view(-1, x1.size(-1))).view(x1.size(0), -1, 1)
        x2_logits = self.y_linear(x2.contiguous().view(-1, x2.size(-1))).view(x2.size(0), 1, -1)
        x1_dot = self.x_dot_linear(x1.contiguous().view(-1, x1.size(-1))).view(x1.size(0), -1, 1).expand_as(x1)
        x1_dot = x1 * x1_dot

        scores = x1_dot.bmm(x2.transpose(1, 2))
        scores += x1_logits.expand_as(scores) + x2_logits.expand_as(scores)
        return scores


class SimilarityWrapper(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='attention', opt={}, dropout=None):
        super(SimilarityWrapper, self).__init__()
        self.score_func_str = opt.get('{}_sim_func'.format(prefix), 'dotproductproject').lower()
        self.score_func = None
        if self.score_func_str == 'dotproduct':
            self.score_func = DotProduct(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'dotproductproject':
            self.score_func = DotProductProject(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'bilinear':
            self.score_func = Bilinear(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'bilinearsum':
            self.score_func = BilinearSum(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'trilinear':
            self.score_func = Trilinear(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        else:
            raise NotImplementedError

    def forward(self, x1, x2):
        scores = self.score_func(x1, x2)
        return scores


class AttentionWrapper(nn.Module):
    def __init__(self, x1_dim, x2_dim, x3_dim=None, prefix='attention', opt={}, dropout=None):
        super(AttentionWrapper, self).__init__()
        self.prefix = prefix
        self.att_dropout = opt.get('{}_att_dropout'.format(self.prefix), 0)
        self.score_func = SimilarityWrapper(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        self.drop_diagonal = opt.get('{}_drop_diagonal'.format(self.prefix), False)
        self.output_size = x2_dim if x3_dim is None else x3_dim

    def forward(self, query, key, value, key_padding_mask=None, return_scores=False):
        logits = self.score_func(query, key)
        key_mask = key_padding_mask.unsqueeze(1).expand_as(logits)
        logits.data.masked_fill_(key_mask.data, -float('inf'))
        if self.drop_diagonal:
            assert logits.size(1) == logits.size(2)
            diag_mask = torch.diag(logits.data.new(logits.size(1)).zero_() + 1).byte().unsqueeze(0).expand_as(logits)
            logits.data.masked_fill_(diag_mask, -float('inf'))

        prob = F.softmax(logits.view(-1, key.size(1)), 1)
        prob = prob.view(-1, query.size(1), key.size(1))
        if self.att_dropout > 0:
            prob = self.dropout(prob)

        if value is None:
            value = key
        attn = prob.bmm(value)
        if return_scores:
            return attn, prob, logits
        else:
            return attn


class LinearSelfAttn(nn.Module):
    """Self attention over a sequence:
    * o_i = softmax(Wx_i) for x_i in X.
    """
    def __init__(self, input_size, dropout=None):
        super(LinearSelfAttn, self).__init__()
        self.linear = nn.Linear(input_size, 1)
        self.dropout = dropout

    def forward(self, x, x_mask):
        x = self.dropout(x)
        x_flat = x.contiguous().view(-1, x.size(-1))
        scores = self.linear(x_flat).view(x.size(0), x.size(1))
        scores.data.masked_fill_(x_mask.data, -float('inf'))
        alpha = F.softmax(scores, 1)
        return alpha.unsqueeze(1).bmm(x).squeeze(1)


class MLPSelfAttn(nn.Module):
    def __init__(self, input_size, opt={}, prefix='attn_sum', dropout=None):
        super(MLPSelfAttn, self).__init__()
        self.prefix = prefix
        self.FC = nn.Linear(input_size, input_size)
        self.linear = nn.Linear(input_size, 1)
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.f = activation(opt.get('{}_activation'.format(self.prefix), 'relu'))
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout
        if self.layer_norm_on:
            self.FC = weight_norm(self.FC)

    def forward(self, x, x_mask):
        x = self.dropout(x)
        x_flat = x.contiguous().view(-1, x.size(-1))
        scores = self.linear(self.f(self.FC(x_flat))).view(x.size(0), x.size(1))
        scores.data.masked_fill_(x_mask.data, -float('inf'))
        alpha = F.softmax(scores)
        return alpha.unsqueeze(1).bmm(x).squeeze(1)


class SelfAttnWrapper(nn.Module):
    def __init__(self, input_size, prefix='attn_sum', opt={}, dropout=None):
        super(SelfAttnWrapper, self).__init__()
        """
        Self att wrapper, support linear and MLP
        """
        attn_type = opt.get('{}_type'.format(prefix), 'linear')
        if attn_type == 'mlp':
            self.att = MLPSelfAttn(input_size, prefix, opt, dropout)
        else:
            self.att = LinearSelfAttn(input_size, dropout)

    def forward(self, x, x_mask):
        return self.att(x, x_mask)


class DeepAttentionWrapper(nn.Module):
    def __init__(self, x1_dim, x2_dim, x3_dims, att_cnt, prefix='deep_att', opt=None, dropout=None):
        super(DeepAttentionWrapper, self).__init__()
        self.opt = {} if opt is None else opt
        self.prefix = prefix
        self.x1_dim = x1_dim
        self.x2_dim = x2_dim
        self.x3_dims = x3_dims

        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout

        self.attn_list = nn.ModuleList()
        for i in range(0, att_cnt):
            if opt['multihead_on']:
                attention = MultiheadAttentionWrapper(self.x1_dim, self.x2_dim, self.x3_dims[i], prefix, opt, dropout=dropout)
            else:
                attention = AttentionWrapper(self.x1_dim, self.x2_dim, self.x3_dims[i], prefix, opt, self.dropout)
            self.attn_list.append(attention)

    def forward(self, x1, x2, x3, x2_mask):
        rvl = []
        for i in range(0, len(x3)):
            hiddens = self.attn_list[i](x1, x2, x3[i], x2_mask)
            rvl.append(hiddens)

        return torch.cat(rvl, 2)


class BilinearFlatSim(nn.Module):
    """A bilinear attention layer over a sequence X w.r.t y:
    * o_i = x_i'Wy for x_i in X.
    """
    def __init__(self, x_size, y_size, opt={}, prefix='seqatt', dropout=None):
        super(BilinearFlatSim, self).__init__()
        self.opt = opt
        self.weight_norm_on = opt.get('{}_weight_norm_on'.format(prefix), False)
        self.linear = nn.Linear(y_size, x_size)
        if self.weight_norm_on:
            self.linear = weight_norm(self.linear)
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout

    def forward(self, x, y, x_mask):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        x = self.dropout(x)
        y = self.dropout(y)

        Wy = self.linear(y)
        xWy = x.bmm(Wy.unsqueeze(2)).squeeze(2)
        xWy.data.masked_fill_(x_mask.data, -float('inf'))
        return xWy


class SimpleFlatSim(nn.Module):
    def __init__(self, x_size, y_size, opt={}, prefix='seqatt', dropout=None):
        super(SimpleFlatSim, self).__init__()
        self.opt = opt
        self.weight_norm_on = opt.get('{}_norm_on'.format(prefix), False)
        self.linear = nn.Linear(y_size + x_size, 1)
        if self.weight_norm_on:
            self.linear = weight_norm(self.linear)
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout

    def forward(self, x, y, x_mask):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        x = self.dropout(x)
        y = self.dropout(y)
        y = y.unsqueeze(1).expand_as(x)
        flat_x = torch.cat([x, y], 2).contiguous().view(x.size(0) * x.size(1), -1)
        flat_scores = self.linear(flat_x)
        scores = flat_scores.contiguous().view(x.size(0), -1)
        scores.data.masked_fill_(x_mask.data, -float('inf'))
        return scores


class FlatSim(nn.Module):
    def __init__(self, x_size, y_size, opt={}, prefix='seqatt', dropout=None):
        super(FlatSim, self).__init__()
        assert x_size == y_size
        self.opt = opt
        self.weight_norm_on = opt.get('{}_weight_norm_on'.format(prefix), False)
        self.linear = nn.Linear(x_size * 3, 1)
        if self.weight_norm_on:
            self.linear = weight_norm(self.linear)
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout

    def forward(self, x, y, x_mask):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        x = self.dropout(x)
        y = self.dropout(y)
        y = y.unsqueeze(1).expand_as(x)

        flat_x = torch.cat([x, y, x * y], 2).contiguous().view(x.size(0) * x.size(1), -1)
        flat_scores = self.linear(flat_x)
        scores = flat_scores.contiguous().view(x.size(0), -1)
        scores.data.masked_fill_(x_mask.data, -float('inf'))

        return scores


class FlatSimV2(nn.Module):
    def __init__(self, x_size, y_size, opt={}, prefix='seqatt', dropout=None):
        super(FlatSimV2, self).__init__()
        assert x_size == y_size
        self.opt = opt
        self.weight_norm_on = opt.get('{}_weight_norm_on'.format(prefix), False)
        self.linear = nn.Linear(x_size * 4, 1)
        if self.weight_norm_on:
            self.linear = weight_norm(self.linear)
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout

    def forward(self, x, y, x_mask):
        """
        x = batch * len * h1
        y = batch * h2
        x_mask = batch * len
        """
        x = self.dropout(x)
        y = self.dropout(y)
        y = y.unsqueeze(1).expand_as(x)

        flat_x = torch.cat([x, y, x * y, torch.abs(x - y)], 2).contiguous().view(x.size(0) * x.size(1), -1)
        flat_scores = self.linear(flat_x)
        scores = flat_scores.contiguous().view(x.size(0), -1)
        scores.data.masked_fill_(x_mask.data, -float('inf'))

        return scores


class FlatSimilarityWrapper(nn.Module):
    def __init__(self, x1_dim, x2_dim, prefix='attention', opt={}, dropout=None):
        super(FlatSimilarityWrapper, self).__init__()
        self.score_func_str = opt.get('{}_att_type'.format(prefix), 'none').lower()
        self.att_dropout = DropoutWrapper(opt.get('{}_att_dropout'.format(prefix), 0))
        self.score_func = None
        if self.score_func_str == 'bilinear':
            self.score_func = BilinearFlatSim(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'simple':
            self.score_func = SimpleFlatSim(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        elif self.score_func_str == 'flatsim':
            self.score_func = FlatSim(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)
        else:
            self.score_func = FlatSimV2(x1_dim, x2_dim, prefix=prefix, opt=opt, dropout=dropout)

    def forward(self, x1, x2, mask):
        scores = self.score_func(x1, x2, mask)
        return scores

class MultiheadAttentionWrapper(nn.Module):
    """Multi-headed attention.
    See "Attention Is All You Need" for more details.
    """
    def __init__(self, query_dim, key_dim, value_dim, prefix='attention', opt={}, dropout=None):
        super().__init__()
        self.prefix = prefix

        self.num_heads = opt.get('{}_head'.format(self.prefix), 1)
        self.dropout = DropoutWrapper(opt.get('{}_dropout'.format(self.prefix), 0)) if dropout is None else dropout

        self.qkv_dim = [query_dim, key_dim, value_dim]
        assert query_dim == key_dim, "query dim must equal with key dim"

        self.hidden_size = opt.get('{}_hidden_size'.format(self.prefix), 64)

        self.proj_on = opt.get('{}_proj_on'.format(prefix), False)
        self.share = opt.get('{}_share'.format(self.prefix), False)
        self.layer_norm_on = opt.get('{}_norm_on'.format(self.prefix), False)
        self.scale_on = opt.get('{}_scale_on'.format(self.prefix), False)

        if self.proj_on:
            self.proj_modules = nn.ModuleList([nn.Linear(dim, self.hidden_size) for dim in self.qkv_dim[0:2]])
            if self.layer_norm_on:
                for proj in self.proj_modules:
                    proj = weight_norm(proj)
            if self.share and self.qkv_dim[0] == self.qkv_dim[1]:
                self.proj_modules[1] = self.proj_modules[0]
            self.f = activation(opt.get('{}_activation'.format(self.prefix), 'relu'))

            self.qkv_head_dim = [self.hidden_size // self.num_heads] * 3
            self.qkv_head_dim[2] = value_dim // self.num_heads
            assert self.qkv_head_dim[0] * self.num_heads == self.hidden_size, "hidden size must be divisible by num_heads"
            assert self.qkv_head_dim[2] * self.num_heads == value_dim, "value size must be divisible by num_heads"

        else:
            self.qkv_head_dim = [emb // self.num_heads for emb in self.qkv_dim]
            #import pdb; pdb.set_trace()
            assert self.qkv_head_dim[0] * self.num_heads == self.qkv_dim[0], "query size must be divisible by num_heads"
            assert self.qkv_head_dim[1] * self.num_heads == self.qkv_dim[1], "key size must be divisible by num_heads"
            assert self.qkv_head_dim[2] * self.num_heads == self.qkv_dim[2], "value size must be divisible by num_heads"

        if self.scale_on:
            self.scaling = self.qkv_head_dim[0]**-0.5
        self.drop_diagonal = opt.get('{}_drop_diagonal'.format(self.prefix), False)
        self.output_size = self.qkv_dim[2]

    def forward(self, query, key, value, key_padding_mask=None):
        query = query.transpose(0, 1)
        key = key.transpose(0, 1)
        value = value.transpose(0, 1)

        tgt_len, bsz, embed_dim = query.size()
        assert embed_dim == self.qkv_dim[0]

        q, k, v = query, key, value
        if self.proj_on:
            if self.dropout:
                q, k = self.dropout(q), self.dropout(k)
            q, k = [self.f(proj(input)) for input, proj in zip([query, key], self.proj_modules)]

        src_len = k.size(0)
        if key_padding_mask is not None:
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.scale_on:
            q *= self.scaling

        q = q.contiguous().view(tgt_len, bsz*self.num_heads, self.qkv_head_dim[0]).transpose(0, 1)
        k = k.contiguous().view(src_len, bsz*self.num_heads, self.qkv_head_dim[1]).transpose(0, 1)
        v = v.contiguous().view(src_len, bsz*self.num_heads, self.qkv_head_dim[2]).transpose(0, 1)

        attn_weights = torch.bmm(q, k.transpose(1, 2))
        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if key_padding_mask is not None:
            # don't attend to padding symbols
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            attn_weights = attn_weights.float().masked_fill(
                key_padding_mask.unsqueeze(1).unsqueeze(2),
                float('-inf'),
            ).type_as(attn_weights)  # FP16 support: cast to float and back
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        if self.drop_diagonal:
            assert attn_weights.size(1) == attn_weights.size(2)
            diag_mask = torch.diag(attn_weights.data.new(attn_weights.size(1)).zero_() + 1).byte().unsqueeze(0).expand_as(attn_weights)
            attn_weights.data.masked_fill_(diag_mask, -float('inf'))

        attn_weights = F.softmax(attn_weights.float(), dim=-1).type_as(attn_weights)
        attn_weights = self.dropout(attn_weights)

        attn = torch.bmm(attn_weights, v)
        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.qkv_head_dim[2]]
        attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, -1)

        # output_shape: Batch * Time * Channel
        attn = attn.transpose(0, 1)

        return attn



import pytest

from gobbli.model.mtdnn import MTDNN


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (max_seq_length)
        ({"max_seq_length": "100"}, TypeError),
        # Bad value (mtdnn_model)
        ({"mtdnn_model": "bert"}, ValueError),
        # OK type (max_seq_length)
        ({"max_seq_length": 100}, None),
        # OK value (mtdnn_model)
        ({"mtdnn_model": "mt-dnn-base"}, None),
        # OK values (both params)
        ({"max_seq_length": 100, "mtdnn_model": "mt-dnn-base"}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        MTDNN(**params)
    else:
        with pytest.raises(exception):
            MTDNN(**params)



import subprocess
from pathlib import Path

import click

INTERACTIVE_DIR = Path(__file__).parent / "interactive"


def _streamlit_run(app_name: str, *args):
    return subprocess.check_call(
        ["streamlit", "run", str(INTERACTIVE_DIR / f"{app_name}.py"), "--", *args]
    )


@click.group()
def main():
    pass


@main.command(
    # Forward the --help argument to the streamlit apps
    "explore",
    context_settings=dict(ignore_unknown_options=True),
    add_help_option=False,
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def main_explore(args):
    _streamlit_run("explore", *args)


@main.command(
    "evaluate",
    context_settings=dict(ignore_unknown_options=True),
    add_help_option=False,
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def main_evaluate(args):
    _streamlit_run("evaluate", *args)


@main.command(
    "explain", context_settings=dict(ignore_unknown_options=True), add_help_option=False
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def main_explain(args):
    _streamlit_run("explain", *args)


if __name__ == "__main__":
    main()



import pytest

from gobbli.model.bert import BERT


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (max_seq_length)
        ({"max_seq_length": "100"}, TypeError),
        # Bad value (bert_model)
        ({"bert_model": "ernie"}, ValueError),
        # OK type (max_seq_length)
        ({"max_seq_length": 100}, None),
        # OK value (bert_model)
        ({"bert_model": "bert-base-uncased"}, None),
        # OK values (both params)
        ({"max_seq_length": 100, "bert_model": "bert-base-uncased"}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        BERT(**params)
    else:
        with pytest.raises(exception):
            BERT(**params)



import os
import argparse
from datetime import datetime
import torch

def predict_config(parser):
    parser.add_argument('--checkpoint', default='d:/model_4.pt')
    parser.add_argument('--fout', type=str, default='mt-dnn-cased-v2.pt')
    return parser

parser = argparse.ArgumentParser()
parser = predict_config(parser)
args = parser.parse_args()

def main():
    opt = vars(args)
    fout = args.fout
    model_path = args.checkpoint
    state_dict = None
    if os.path.exists(model_path):
        state_dict = torch.load(model_path)
        config = state_dict['config']
        opt.update(config)
        if state_dict['config']['ema_opt'] > 0:
            new_state_dict = {'state': state_dict['ema'], 'config': state_dict['config']}
        else:
            new_state_dict = {'state': state_dict['state'], 'config': state_dict['config']}
        old_state_dict = {}
        for key, val in new_state_dict['state'].items():
            prefix = key.split('.')[0]
            if prefix == 'scoring_list':
                continue
            old_state_dict[key] = val
        my_config = {}
        my_config['vocab_size'] = config['vocab_size']
        my_config['hidden_size'] = config['hidden_size']
        my_config['num_hidden_layers'] = config['num_hidden_layers']
        my_config['num_attention_heads'] = config['num_attention_heads']
        my_config['hidden_act'] = config['hidden_act']
        my_config['intermediate_size'] = config['intermediate_size']
        my_config['hidden_dropout_prob'] = config['hidden_dropout_prob']
        my_config['attention_probs_dropout_prob'] = config['attention_probs_dropout_prob']
        my_config['max_position_embeddings'] = config['max_position_embeddings']
        my_config['type_vocab_size'] = config['type_vocab_size']
        my_config['initializer_range'] = config['initializer_range']
        state_dict = {'state': old_state_dict, 'config': my_config}
        torch.save(state_dict, args.fout)

if __name__ == '__main__':
    main()


from pathlib import Path
from typing import Any

import pytest

from gobbli.dataset.base import BaseDataset
from gobbli.experiment.base import BaseExperiment
from gobbli.model.base import BaseModel
from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText, FastTextCheckpoint
from gobbli.model.mtdnn import MTDNN
from gobbli.model.transformer import Transformer
from gobbli.util import gobbli_dir


def skip_if_no_gpu(config):
    if not config.option.use_gpu:
        pytest.skip("needs --use-gpu option to run")


def skip_if_low_resource(config):
    """
    Used when a test involves a large amount of CPU, memory, etc,
    and the user has indicated we're running in a resource-limited
    environment.
    """
    if config.option.low_resource:
        pytest.skip("skipping large test due to --low-resource option")


# TODO can we write a type declaration to indicate that args
# should be classes derived from BaseModel?
def model_test_dir(model_cls: Any) -> Path:
    """
    Return a directory to be used for models of the passed type.
    Helpful when the user wants data to be persisted so weights don't
    have to be reloaded for each test run.
    """
    return gobbli_dir() / "model_test" / model_cls.__name__


def validate_checkpoint(model_cls: Any, checkpoint: Path):
    """
    Use assertions to validate a given checkpoint depending on which kind of
    model created it.
    """
    if model_cls == BERT:
        # The checkpoint isn't an actual file, but it should have an associated metadata file
        assert Path(f"{str(checkpoint)}.meta").is_file()
    elif model_cls == MTDNN:
        # The checkpoint is a single file
        assert checkpoint.is_file()
    elif model_cls == FastText:
        # The checkpoint has a couple components
        fasttext_checkpoint = FastTextCheckpoint(checkpoint)
        assert fasttext_checkpoint.model.exists()
        assert fasttext_checkpoint.vectors.exists()
    elif model_cls == Transformer:
        assert checkpoint.is_dir()


class MockDataset(BaseDataset):
    """
    A minimal dataset derived from BaseDataset for testing the
    ABC's logic.
    """

    X_TRAIN_VALID = ["train1", "train2", "train3", "train4"]
    Y_TRAIN_VALID = ["0", "1", "0", "1"]

    X_TEST = ["test1", "test2"]
    Y_TEST = ["1", "0"]

    def __init__(self, *args, **kwargs):
        self._built = False
        self._build_count = 0

    def _is_built(self) -> bool:
        return self._built

    def _build(self):
        self._build_count += 1
        self._built = True

    def X_train(self):
        return MockDataset.X_TRAIN_VALID

    def y_train(self):
        return MockDataset.Y_TRAIN_VALID

    def X_test(self):
        return MockDataset.X_TEST

    def y_test(self):
        return MockDataset.Y_TEST


class MockModel(BaseModel):
    """
    A minimal model derived from BaseModel for testing the
    ABC's logic.
    """

    def init(self, params):
        self.params = params

    def _build(self):
        pass


class MockExperiment(BaseExperiment):
    """
    A minimal experiment derived from BaseExperiment for testing the
    ABC's logic.
    """



import pytest

from gobbli.model.spacy import SpaCyModel


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (dropout)
        ({"dropout": "100"}, TypeError),
        # OK type (dropout)
        ({"dropout": 0.3}, None),
        # Bad type (full_pipeline)
        ({"full_pipeline": 1}, TypeError),
        # OK type (full_pipeline)
        ({"full_pipeline": True}, None),
        # OK types (all params)
        ({"full_pipeline": True, "dropout": 0.3}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        SpaCyModel(**params)
    else:
        with pytest.raises(exception):
            SpaCyModel(**params)



import json
from typing import List, Tuple

import pandas as pd

from gobbli.dataset.base import BaseDataset
from gobbli.util import download_archive


class MovieSummaryDataset(BaseDataset):
    """
    gobbli Dataset for the CMU Movie Summary dataset, framed as a multilabel
    classification problem predicting movie genres from plot summaries.

    http://www.cs.cmu.edu/~ark/personas/
    """

    PLOT_SUMMARIES_FILE = "MovieSummaries/plot_summaries.txt"
    METADATA_FILE = "MovieSummaries/movie.metadata.tsv"
    TRAIN_PCT = 0.8

    def _build(self):
        data_dir = self.data_dir()
        data_dir.mkdir(exist_ok=True, parents=True)

        download_archive(
            "http://www.cs.cmu.edu/~ark/personas/data/MovieSummaries.tar.gz", data_dir
        )

    @staticmethod
    def _make_multilabels(genres: pd.Series) -> List[List[str]]:
        return [list(json.loads(g).values()) for g in genres]

    def _is_built(self) -> bool:
        data_dir = self.data_dir()
        return (data_dir / MovieSummaryDataset.PLOT_SUMMARIES_FILE).exists() and (
            data_dir / MovieSummaryDataset.METADATA_FILE
        ).exists()

    def _get_source_df_split(self) -> Tuple[pd.DataFrame, int]:
        if not hasattr(self, "_source_df"):
            data_dir = self.data_dir()
            plot_df = pd.read_csv(
                data_dir / MovieSummaryDataset.PLOT_SUMMARIES_FILE,
                delimiter="\t",
                index_col=0,
                header=None,
                names=["wiki_id", "plot"],
            )

            meta_df = pd.read_csv(
                data_dir / MovieSummaryDataset.METADATA_FILE,
                delimiter="\t",
                index_col=0,
                header=None,
                names=[
                    "wiki_id",
                    "freebase_id",
                    "name",
                    "release_date",
                    "revenue",
                    "runtime",
                    "languages",
                    "countries",
                    "genres",
                ],
            )

            self._source_df = plot_df.join(meta_df, how="inner")[
                ["plot", "genres"]
            ].sort_index()

        return (
            self._source_df,
            int(len(self._source_df) * MovieSummaryDataset.TRAIN_PCT),
        )

    def X_train(self):
        source_df, split_ndx = self._get_source_df_split()
        return source_df["plot"].tolist()[:split_ndx]

    def y_train(self):
        source_df, split_ndx = self._get_source_df_split()
        return MovieSummaryDataset._make_multilabels(source_df["genres"][:split_ndx])

    def X_test(self):
        source_df, split_ndx = self._get_source_df_split()
        return source_df["plot"].tolist()[split_ndx:]

    def y_test(self):
        source_df, split_ndx = self._get_source_df_split()
        return MovieSummaryDataset._make_multilabels(source_df["genres"][split_ndx:])



import pytest

from gobbli.augment.marian import MarianMT
from gobbli.test.util import model_test_dir


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (batch_size)
        ({"batch_size": 2.5}, TypeError),
        # Bad type (target_languages)
        ({"target_languages": "english"}, TypeError),
        # Bad value (batch_size)
        ({"batch_size": 0}, ValueError),
        # Bad value (target_languages)
        ({"target_languages": ["not a language"]}, ValueError),
        # Bad value, one OK value (target_languages)
        ({"target_languages": ["french", "not a language"]}, ValueError),
        # OK values
        ({"batch_size": 16, "target_languages": ["russian", "french"]}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        MarianMT(**params)
    else:
        with pytest.raises(exception):
            MarianMT(**params)


def test_marianmt_augment(model_gpu_config, gobbli_dir):
    # Don't go overboard with the languages here, since each
    # one requires a separate model (few hundred MB) to be downloaded
    target_languages = ["russian", "french"]
    model = MarianMT(
        data_dir=model_test_dir(MarianMT),
        load_existing=True,
        target_languages=target_languages,
        **model_gpu_config,
    )
    model.build()

    # Can't augment more times than target languages
    invalid_num_times = len(target_languages) + 1
    with pytest.raises(ValueError):
        model.augment(["This is a test."], times=invalid_num_times)

    valid_num_times = len(target_languages)
    new_texts = model.augment(["This is a test."], times=valid_num_times)
    assert len(new_texts) == valid_num_times



import argparse
import ast
import json
import random
from pathlib import Path

import numpy as np
import pandas as pd
import spacy
from spacy.gold import GoldParse
from spacy.util import minibatch

import torch
from spacy_transformers import TransformersLanguage
from spacy_transformers.util import PIPES, cyclic_triangular_rate


def is_transformer(nlp):
    """
    Determine whether the given spacy language instance is backed
    by a transformer model or a regular spaCy model.
    """
    return isinstance(nlp, TransformersLanguage)


def read_unique_labels(labels_path):
    """
    Read the list of unique labels from the given path.
    """
    return labels_path.read_text(encoding="utf-8").split("\n")


def read_data(file_path, has_labels):
    """
    Read some text data with optional labels from the given path.
    Return a 2-tuple of texts and labels (if any).
    """
    df = pd.read_csv(file_path, sep="\t", dtype="str", keep_default_na=False)
    X = df["Text"].tolist()
    y = None
    if has_labels:
        # Since spaCy defaults to multilabel paradigm, read in all data in that
        # format -- assume labels are a list
        y = df["Label"].apply(lambda labels: set(ast.literal_eval(labels))).tolist()
    return X, y


def spacy_format_labels(ys, labels):
    """Convert a list of labels to the format spaCy expects for model training."""
    return [{l: int(l in y) for l in labels} for y in ys]


def evaluate(tokenizer, nlp, valid_data, labels):
    """Evaluate model performance on a test dataset."""
    texts, cats = zip(*valid_data)

    golds = []
    # Use the model's ops module
    # to make sure this is compatible with GPU (cupy array)
    # or without (numpy array)
    scores = np.zeros((len(cats), len(labels)), dtype="f")
    if is_transformer(nlp):
        textcat = nlp.get_pipe(PIPES.textcat)
    else:
        textcat = nlp.get_pipe("textcat")
    scores = textcat.model.ops.asarray(scores)

    num_correct = 0
    for i, doc in enumerate(nlp.pipe(texts)):
        gold_cats = cats[i]["cats"]
        for j, (label, score) in enumerate(doc.cats.items()):
            if label not in gold_cats:
                raise ValueError(f"Prediction for unexpected label: {label}")

            scores[i, j] = score

            doc_prediction = score > 0.5
            if doc_prediction == bool(gold_cats[label]):
                num_correct += 1

        golds.append(GoldParse(doc, cats=gold_cats))

    accuracy = num_correct / ((len(texts) * len(labels)) + 1e-8)
    loss, _ = textcat.get_loss(texts, golds, scores)

    return accuracy, loss


def train(
    *,
    input_dir,
    output_dir,
    nlp,
    architecture,
    train_batch_size,
    num_train_epochs,
    labels,
    dropout,
    disabled_components,
    multilabel,
):
    """
    Train the TextCategorizer component of the passed pipeline on the given data.
    Return training/validation metrics and save the model to the specified output directory.
    Make sure to restore any disabled pipeline components before saving so we can reuse the
    saved checkpoint however we need to.
    """
    if is_transformer(nlp):
        textcat_pipe_name = PIPES.textcat
        textcat = nlp.create_pipe(
            textcat_pipe_name,
            config={
                "architecture": "softmax_last_hidden",
                "exclusive_classes": not multilabel,
                # We get an error about token_vector_width being unset if it isn't set
                # explicitly here.  We can't set it to an arbitrary value, either.  It must
                # be set based on the model
                "token_vector_width": nlp.get_pipe(PIPES.tok2vec).model.nO,
            },
        )
    else:
        textcat_pipe_name = "textcat"
        textcat = nlp.create_pipe(
            textcat_pipe_name,
            config={"exclusive_classes": not multilabel, "architecture": architecture},
        )
    nlp.add_pipe(textcat, last=True)

    for label in labels:
        textcat.add_label(label)

    X_train, y_train = read_data(input_dir / "train.tsv", True)
    X_valid, y_valid = read_data(input_dir / "dev.tsv", True)

    train_labels = spacy_format_labels(y_train, labels)
    valid_labels = spacy_format_labels(y_valid, labels)

    train_data = list(zip(X_train, [{"cats": cats} for cats in train_labels]))
    valid_data = list(zip(X_valid, [{"cats": cats} for cats in valid_labels]))

    with nlp.disable_pipes(*disabled_components):
        if is_transformer(nlp):
            optimizer = nlp.resume_training()
            optimizer.alpha = 0.001
            optimizer.trf_weight_decay = 0.005
            optimizer.L2 = 0.0
            learn_rate = 2e-5
            learn_rates = cyclic_triangular_rate(
                learn_rate / 3, learn_rate * 3, 2 * len(train_data) // train_batch_size
            )
        else:
            optimizer = nlp.begin_training()
        for i in range(num_train_epochs):
            losses = {}
            random.shuffle(train_data)
            batches = minibatch(train_data, train_batch_size)
            for batch in batches:
                texts, annotations = zip(*batch)
                if is_transformer(nlp):
                    optimizer.trf_lr = next(learn_rates)

                nlp.update(
                    texts, annotations, sgd=optimizer, drop=dropout, losses=losses
                )

            with textcat.model.use_params(optimizer.averages):
                accuracy, valid_loss = evaluate(nlp.tokenizer, nlp, valid_data, labels)
                train_loss = losses[textcat_pipe_name]
                print(
                    f"Iter {i}\tTrain Loss: {train_loss:.3f}\tValid Loss: {valid_loss:.3f}\tAccuracy: {accuracy:.3f}"
                )

    checkpoint_dir = output_dir / "checkpoint"
    checkpoint_dir.mkdir(exist_ok=True, parents=True)

    with nlp.use_params(optimizer.averages):
        nlp.to_disk(checkpoint_dir)

    metrics = {
        "valid_accuracy": accuracy,
        "mean_train_loss": losses[textcat_pipe_name] / len(X_train),
        "mean_valid_loss": valid_loss / len(X_valid),
    }

    with open(output_dir / "valid_results.json", "w") as f:
        json.dump(metrics, f)


def predict(*, input_dir, output_dir, nlp, labels, disabled_components):
    """
    Generate predictions for the given dataset using the TextCategorizer component
    of the passed pipeline.
    """
    X_test, _ = read_data(input_dir / "test.tsv", False)

    pred_probas = []
    with nlp.disable_pipes(*disabled_components):
        for doc in nlp.pipe(X_test):
            pred_probas.append({label: doc.cats.get(label, 0.0) for label in labels})
    df = pd.DataFrame(pred_probas)
    df.to_csv(output_dir / "test_results.tsv", index=False, sep="\t")


def embed(*, input_dir, output_dir, nlp, embed_pooling, disabled_components):
    """
    Generate embeddings for the given dataset using the vectors from the
    passed pipeline.
    """
    embeddings = []
    X_embed, _ = read_data(input_dir / "input.tsv", False)

    with nlp.disable_pipes(*disabled_components):
        with open(output_dir / "embeddings.jsonl", "w") as f:
            for doc in nlp.pipe(X_embed):
                if embed_pooling == "mean":
                    row_json = {"embedding": doc.vector.tolist()}
                elif embed_pooling == "none":
                    embeddings = []
                    tokens = []
                    for tok in doc:
                        embeddings.append(tok.vector.tolist())
                        tokens.append(tok.text)
                    row_json = {"embedding": embeddings, "tokens": tokens}
                else:
                    raise ValueError(f"Unsupported pooling type: {embed_pooling}")

                f.write(f"{json.dumps(row_json)}\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "mode",
        choices=["train", "predict", "embed"],
        help="Action to perform with the model. Train: read in a training "
        "and validation dataset, train on training and evaluate the model "
        "on validation dataset. Predict: output predictions on a test dataset. "
        "Embed: return embeddings for a test dataset.",
    )
    parser.add_argument(
        "--input-dir",
        required=True,
        help="Directory containing input files.  The exact files needed vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--output-dir",
        required=True,
        help="Directory containing output files.  The exact files created vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--cache-dir",
        required=True,
        help="Directory to use for caching spaCy downloads.",
    )
    parser.add_argument(
        "--model",
        required=True,
        help="SpaCy language model to use.  This could be "
        "either the name of a stock spaCy model (in which case it's assumed to already "
        "be installed via pip) or a path to a custom spaCy model on disk.",
    )
    parser.add_argument(
        "--architecture",
        required=True,
        help="Architecture for the spaCy TextCategorizer.",
    )
    parser.add_argument(
        "--multilabel",
        action="store_true",
        help="If True, model will train in a multilabel context (i.e. it's allowed to make multiple "
        "class predictions for each observation).",
    )
    parser.add_argument(
        "--full-pipeline",
        action="store_true",
        help="If passed, use the full spaCy language pipeline (including tagging, "
        "parsing, and named entity recognition) for the TextCategorizer model used in "
        "training and prediction.  This makes training/prediction much slower but theoretically "
        "provides more information to the model.",
    )
    parser.add_argument(
        "--train-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for training.",
    )
    parser.add_argument(
        "--embed-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for embedding.",
    )
    parser.add_argument(
        "--num-train-epochs",
        type=int,
        default=3,
        help="Number of epochs to run training on.",
    )
    parser.add_argument(
        "--dropout", type=float, default=0.2, help="Dropout proportion for training."
    )
    parser.add_argument(
        "--embed-pooling",
        choices=["mean", "none"],
        default="mean",
        help="Pooling strategy to use for combining embeddings.  If None, embeddings for "
        "each token will be returned along with the token.  Ignored if method != 'embed'.",
    )

    args = parser.parse_args()

    input_dir = Path(args.input_dir)
    output_dir = Path(args.output_dir)

    using_gpu = spacy.prefer_gpu()
    if using_gpu:
        torch.set_default_tensor_type("torch.cuda.FloatTensor")

    device = "gpu" if using_gpu else "cpu"
    print(f"Using device: {device}")

    print("Initializing spaCy model...")
    print(f"  Language Model: {args.model}")

    nlp = spacy.load(args.model)
    if not is_transformer(nlp):
        print(f"  TextCategorizer Architecture: {args.architecture}")

    model_name = nlp.meta.get("name", "")

    print(f"Model '{model_name}' loaded.")

    disabled_components = set()

    if args.mode == "embed":
        # Don't need the text categorizer (if present) for embeddings
        for textcat_pipe in ("textcat", "trf_textcat"):
            if nlp.has_pipe(textcat_pipe):
                disabled_components.add(textcat_pipe)

        if model_name.endswith("sm"):
            # No vectors available for small models -- we need to enable the
            # other pipeline components to provide tensors, which aren't as good as vectors
            # but will suffice in a pinch
            pass
        else:
            # If vectors are available, disable everything, since we just want the vectors
            for component in ("tagger", "parser", "ner"):
                if nlp.has_pipe(component):
                    disabled_components.add(component)

    elif args.mode in ("train", "predict"):
        if args.full_pipeline:
            # Enable all parsing components to provide maximum information to the
            # text categorization model
            pass
        else:
            # Otherwise, disable everything that isn't part of the text categorizer model
            for component in ("tagger", "parser", "ner"):
                if nlp.has_pipe(component):
                    disabled_components.add(component)

    # We need a list of labels for training or prediction so we know what the
    # output shape of the model should be
    labels = None
    if args.mode in ("train", "predict"):
        labels_file = input_dir / "labels.tsv"
        labels = read_unique_labels(input_dir / "labels.tsv")
        num_labels = len(labels)
        print(f"Inferred number of labels: {num_labels}")

    if args.mode == "train":
        train(
            input_dir=input_dir,
            output_dir=output_dir,
            nlp=nlp,
            architecture=args.architecture,
            labels=labels,
            train_batch_size=args.train_batch_size,
            num_train_epochs=args.num_train_epochs,
            dropout=args.dropout,
            disabled_components=disabled_components,
            multilabel=args.multilabel,
        )

    elif args.mode == "predict":
        predict(
            input_dir=input_dir,
            output_dir=output_dir,
            nlp=nlp,
            labels=labels,
            disabled_components=disabled_components,
        )

    elif args.mode == "embed":
        embed(
            input_dir=input_dir,
            output_dir=output_dir,
            nlp=nlp,
            embed_pooling=args.embed_pooling,
            disabled_components=disabled_components,
        )

    else:
        raise ValueError(f"invalid mode: {args.mode}")



import argparse

from transformers import MarianMTModel, MarianTokenizer


def batch_list(l, batch_size):
    for i in range(0, len(l), batch_size):
        yield l[i : i + batch_size]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "input_file",
        help="Input file containing line-delimited texts. "
        "Note multi-language Marian models require each text to be prefixed with '>>id<<', "
        "where 'id' is the ID of a target language for translation.  This script assumes "
        "the >>id<< prefix is correctly placed (or not) depending on the type of model.",
    )
    parser.add_argument(
        "output_file",
        help="Output file, where line-delimited generated texts will be written.",
    )
    parser.add_argument(
        "--marian-model",
        help="Marian model to use. "
        "Can be one of the pretrained names supported by transformers, in which case "
        "the pretrained weights will be downloaded. "
        "Anything else supported by transformers should work as well. "
        "Note an inverse model is required, so a custom model will need a corresponding "
        "inverse to translate in the other direction.",
        required=True,
    )
    parser.add_argument(
        "--marian-inverse-model",
        help="Inverse Marian model to use. "
        "Should be the opposite of --marian-model.  Ex: "
        "--marian-model=Helsinki-NLP/opus-mt-en-ROMANCE "
        "--marian-inverse-model=Helsinki-NLP/opus-mt-ROMANCE-en",
        required=True,
    )
    parser.add_argument(
        "--batch-size",
        help="Number of documents to run through the model at once.",
        type=int,
        default=32,
    )
    parser.add_argument(
        "--cache-dir",
        help="Directory to use as the cache for transformers downloads.",
        default=None,
    )
    parser.add_argument(
        "--device",
        default="cpu",
        help="PyTorch name for the device to use when running the model. Default: %(default)s",
    )

    args = parser.parse_args()

    if args.batch_size < 1:
        raise ValueError("batch_size must be >= 1")

    tokenizer = MarianTokenizer.from_pretrained(
        args.marian_model, cache_dir=args.cache_dir
    )
    if tokenizer is None:
        raise ValueError("Failed to acquire tokenizer")
    model = MarianMTModel.from_pretrained(args.marian_model, cache_dir=args.cache_dir)
    if model is None:
        raise ValueError("Failed to acquire model")

    model = model.to(args.device)
    model.eval()

    inv_tokenizer = MarianTokenizer.from_pretrained(
        args.marian_inverse_model, cache_dir=args.cache_dir
    )
    if inv_tokenizer is None:
        raise ValueError("Failed to acquire inverse tokenizer")
    inv_model = MarianMTModel.from_pretrained(
        args.marian_inverse_model, cache_dir=args.cache_dir
    )
    if inv_model is None:
        raise ValueError("Failed to acquire inverse model")

    inv_model = inv_model.to(args.device)
    inv_model.eval()

    with open(args.input_file, "r", encoding="utf-8") as f_in:
        with open(args.output_file, "w", encoding="utf-8") as f_out:
            batches = list(batch_list(f_in.readlines(), args.batch_size))
            for batch_id, batch in enumerate(batches):
                # Explicit max length needed to make the tokenizer cut off at the max length
                # Otherwise we get potential index errors running the texts through the models
                translated = model.generate(
                    **tokenizer.prepare_translation_batch(
                        batch, max_length=tokenizer.model_max_length
                    ).to(args.device)
                )
                translated_texts = [
                    tokenizer.decode(t, skip_special_tokens=True) for t in translated
                ]

                backtranslated = inv_model.generate(
                    **inv_tokenizer.prepare_translation_batch(
                        translated_texts, max_length=inv_tokenizer.model_max_length
                    ).to(args.device)
                )
                backtranslated_texts = [
                    inv_tokenizer.decode(t, skip_special_tokens=True)
                    for t in backtranslated
                ]
                # Write text to the output file
                # Escape any embedded newlines
                # Make sure not to write an extra newline after the last row
                for i, row in enumerate(backtranslated_texts):
                    f_out.write(row.replace("\n", " "))

                    # Write a newline except for the last row on the last
                    # time through.  We'd get an empty line at the end of the
                    # file otherwise
                    if not (
                        i == len(translated_texts) - 1 and batch_id == len(batches) - 1
                    ):
                        f_out.write("\n")



from pathlib import Path

import gensim
import pytest

from gobbli.augment.word2vec import _WORD2VEC_MODEL_ARCHIVES, WORD2VEC_MODELS, Word2Vec
from gobbli.util import TokenizeMethod


def test_word2vec_models():
    for model_name, (archive_name, filename) in WORD2VEC_MODELS.items():
        assert archive_name in _WORD2VEC_MODEL_ARCHIVES


_TEST_WORD2VEC_VECTORS = """2 3
test 0 1 1
test2 1 0 1
"""


def make_weights_file(tmpdir):
    weights_file = Path(tmpdir) / "word2vec.txt"
    weights_file.write_text(_TEST_WORD2VEC_VECTORS)
    return weights_file


def test_word2vec_init(tmpdir):
    # Path corresponding to existing weights file should be loaded
    weights_file = make_weights_file(tmpdir)
    Word2Vec(weights_file)

    # Existing model should be loaded
    m = gensim.models.KeyedVectors.load_word2vec_format(str(weights_file))
    Word2Vec(m)

    # String corresponding to one of the pretrained models should download and
    # use the weights
    Word2Vec("fasttext-simple")

    # Anything else should error
    with pytest.raises(TypeError):
        Word2Vec(None)


@pytest.mark.parametrize(
    "n_similar,diversity,tokenizer,exception_cls",
    [
        # wrong type n_similar
        (5.5, 1.0, TokenizeMethod.SPLIT, TypeError),
        # wrong type diversity
        (5, 1, TokenizeMethod.SPLIT, TypeError),
        # wrong type tokenizer
        (5, 0.8, 1, TypeError),
        # bad value n_similar
        (0, 1.0, TokenizeMethod.SPLIT, ValueError),
        # bad value diversity (<= 0)
        (5, 0.0, TokenizeMethod.SPLIT, ValueError),
        # bad value diversity (> 1)
        (5, 1.1, TokenizeMethod.SPLIT, ValueError),
        # bad value str tokenizer (no match)
        (5, 0.8, "no such tokenizer", KeyError),
        # ok (enum tokenizer)
        (5, 0.8, TokenizeMethod.SPLIT, None),
        # ok (callable tokenizer)
        (5, 0.8, lambda s: s.split(), None),
        # ok (str tokenizer)
        (5, 0.8, "SPLIT", None),
    ],
)
def test_word2vec_kwargs(tmpdir, n_similar, diversity, tokenizer, exception_cls):
    weights_file = make_weights_file(tmpdir)
    kwargs = {"n_similar": n_similar, "diversity": diversity, "tokenizer": tokenizer}

    if exception_cls is None:
        Word2Vec(weights_file, **kwargs)
    else:
        with pytest.raises(exception_cls):
            Word2Vec(weights_file, **kwargs)


def test_word2vec_replace(tmpdir):
    weights_file = make_weights_file(tmpdir)

    m = Word2Vec(weights_file)

    # we should find a replacement for a word in the vocabulary
    assert m._maybe_replace_token("test") == "test2"

    # out of vocabulary words shouldn't be replaced
    assert m._maybe_replace_token("bad_token") == "bad_token"



import numpy as np

import torch
import torch.nn.functional as F
from torch.autograd import Variable

from .model import MTDNNModel


class GobbliMTDNNModel(MTDNNModel):
    def update(self, input_ids, token_type_ids, attention_mask, labels):
        self.network.train()
        if self.config['cuda']:
            labels = labels.cuda(async=True)

        y = Variable(labels, requires_grad=False)
        logits = self.mnetwork(input_ids, token_type_ids, attention_mask, task_id=0)
        loss = F.cross_entropy(logits, y)

        self.train_loss.update(loss.item(), logits.size(0))
        self.optimizer.zero_grad()

        loss.backward()
        if self.config['global_grad_clipping'] > 0:
            torch.nn.utils.clip_grad_norm_(self.network.parameters(),
                                           self.config['global_grad_clipping'])

        self.optimizer.step()
        self.updates += 1
        self.update_ema()

    def predict(self, input_ids, token_type_ids, attention_mask):
        self.network.eval()
        score = self.mnetwork(input_ids, token_type_ids, attention_mask, task_id=0)
        score = F.softmax(score, dim=1).data.cpu()
        predict = np.argmax(score.numpy(), axis=1).tolist()
        return score, predict



import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

from gobbli.augment.base import BaseAugment
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.util import assert_type, escape_line_delimited_texts


class MarianMT(BaseModel, BaseAugment):
    """
    Backtranslation-based data augmenter using the Marian Neural Machine Translation
    model.  Translates English text to one of several languages and back to obtain
    similar texts for training.
    """

    _BUILD_PATH = Path(__file__).parent

    _INPUT_FILE = "input.txt"
    _OUTPUT_FILE = "output.txt"

    _CONTAINER_CACHE_DIR = Path("/cache")

    # Hardcoded list based on the languages available in various models
    # Ignore models that translate multiple languages because they have
    # a different API (texts require a >>lang_code<< prelude describing
    # which language to translate to)
    # https://huggingface.co/models?search=opus-mt-en&sort=alphabetical
    # The below mapping was reconstructed manually using various resources
    # including ISO language codes, Google Translate auto-detect,
    # and the JWS docs from Opus: http://opus.nlpl.eu/JW300.php (see languages.json)
    LANGUAGE_CODE_MAPPING = {
        "afrikaans": "af",
        "central-bikol": "bcl",
        "bemba": "bem",
        "berber": "ber",
        "bulgarian": "bg",
        "bislama": "bi",
        # BZS stands for Brazilian Sign Language, but the text
        # looks like Portugese, and there's no other model for Portugese
        "portugese": "bzs",
        "catalan": "ca",
        "cebuano": "ceb",
        "chuukese": "chk",
        "seychelles-creole": "crs",
        "czech": "cs",
        "welsh": "cy",
        "danish": "da",
        "german": "de",
        "ewe": "ee",
        "efik": "efi",
        "esperanto": "eo",
        "estonian": "et",
        "basque": "eu",
        "finnish": "fi",
        "fijian": "fj",
        "french": "fr",
        "irish": "ga",
        "ga": "gaa",
        "gilbertese": "gil",
        "galician": "gl",
        "gun": "guw",
        "manx": "gv",
        "hausa": "ha",
        "hiligaynon": "hil",
        "hiri-motu": "ho",
        "haitian": "ht",
        "hungarian": "hu",
        "indonesian": "id",
        "igbo": "ig",
        "iloko": "ilo",
        "icelandic": "is",
        "isoko": "iso",
        "italian": "it",
        "japanese": "jap",
        "kongo": "kg",
        "kuanyama": "kj",
        "kikaonde": "kqn",
        "kwangali": "kwn",
        "kikongo": "kwy",
        "luganda": "lg",
        "lingala": "ln",
        "silozi": "loz",
        "kiluba": "lu",
        "tshiluba": "lua",
        "luvale": "lue",
        "lunda": "lun",
        "luo": "luo",
        "mizo": "lus",
        "mauritian-creole": "mfe",
        "malagasy": "mg",
        "marshallese": "mh",
        "macedonian": "mk",
        "malayalam": "ml",
        "moore": "mos",
        "marathi": "mr",
        "maltese": "mt",
        "ndonga": "ng",
        "niuean": "niu",
        "dutch": "nl",
        "sepedi": "nso",
        "chichewa": "ny",
        "nyaneka": "nyk",
        "oromo": "om",
        "pangasinan": "pag",
        "papiamento": "pap",
        "solomon-islands-pidgin": "pis",
        "ponapean": "pon",
        "uruund": "rnd",
        "russian": "ru",
        "kirundi": "run",
        "kinyarwanda": "rw",
        "sango": "sg",
        "slovak": "sk",
        "samoan": "sm",
        "shona": "sn",
        "albanian": "sq",
        "swati": "ss",
        "sesotho-lesotho": "st",
        "swedish": "sv",
        "swahili-congo": "swc",
        "tigrinya": "ti",
        "tiv": "tiv",
        "tagalog": "tl",
        "otetela": "tll",
        "setswana": "tn",
        "tongan": "to",
        "chitonga": "toi",
        "tok-pisin": "tpi",
        "tsonga": "ts",
        "tuvaluan": "tvl",
        "ukrainian": "uk",
        "umbundu": "umb",
        "xhosa": "xh",
    }

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        MarianMT parameters:

        - ``batch_size``: Number of documents to run through the Marian model at once.
        - ``target_languages``: List of target languages to translate texts to and back.
          See :attr:`MarianMT.ALL_TARGET_LANGUAGES` for a full list of possible values. You may
          only augment texts up to the number of languages specified, since each language
          will be used at most once.  So if you want to augment 5 times, you need to specify
          at least 5 languages when initializing the model.
        """
        self.batch_size = 32
        # Current default - top 5 lanugages in Wikipedia which are also available
        # in the list of target languages
        # https://en.wikipedia.org/wiki/List_of_Wikipedias#List
        self.target_languages = ["french", "german", "japanese", "russian", "italian"]

        for name, value in params.items():
            if name == "batch_size":
                assert_type(name, value, int)
                if value < 1:
                    raise ValueError("batch_size must be >= 1")
                self.batch_size = value
            elif name == "target_languages":
                assert_type(name, value, list)
                for target in value:
                    if target not in MarianMT.LANGUAGE_CODE_MAPPING:
                        raise ValueError(
                            f"invalid target language '{target}'. Valid values are "
                            f"{list(MarianMT.LANGUAGE_CODE_MAPPING.keys())}"
                        )
                self.target_languages = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the Marian container.
        """
        return f"gobbli-marian-nmt"

    @classmethod
    def marian_model(cls, language: str) -> str:
        """
        Returns:
          Name of the Marian MT model to use to translate English
          to the passed language.
        """
        return f"Helsinki-NLP/opus-mt-en-{cls.LANGUAGE_CODE_MAPPING[language]}"

    @classmethod
    def marian_inverse_model(cls, language: str) -> str:
        """
        Returns:
          Name of the Marian MT model to use to translate the passed language
          back to English.
        """
        return f"Helsinki-NLP/opus-mt-{cls.LANGUAGE_CODE_MAPPING[language]}-en"

    def _build(self):
        self.docker_client.images.build(
            path=str(MarianMT._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    def _write_input(self, X: List[str], context: ContainerTaskContext):
        """
        Write the user input to a file for the container to read.
        """
        input_path = context.host_input_dir / MarianMT._INPUT_FILE
        input_path.write_text(escape_line_delimited_texts(X))

    def _read_output(self, context: ContainerTaskContext) -> List[str]:
        """
        Read generated text output to a file from the container.
        """
        output_path = context.host_output_dir / MarianMT._OUTPUT_FILE
        return output_path.read_text().split("\n")

    @property
    def host_cache_dir(self):
        """
        Directory to be used for downloaded transformers files.
        Should be the same across all instances of the class, since these are
        generally static model weights/config files that can be reused.
        """
        cache_dir = MarianMT.model_class_dir() / "cache"
        cache_dir.mkdir(exist_ok=True, parents=True)
        return cache_dir

    def augment(
        self, X: List[str], times: Optional[int] = None, p: float = None
    ) -> List[str]:
        if times is None:
            times = len(self.target_languages)

        if times > len(self.target_languages):
            raise ValueError(
                "MarianMT was asked to augment {len(times)} times but was only initialized with "
                "{len(self.target_languages)} target languages.  You must specify at least as "
                "many target languages as the number of times you'd like to augment."
            )
        if p is not None:
            warnings.warn(
                "MarianMT doesn't replace text at the token level, so the 'p' parameter "
                "will be ignored."
            )

        context = ContainerTaskContext(self.data_dir())

        self._write_input(X, context)

        # Determine which device to use for augmentation
        device = "cpu"
        if self.use_gpu:
            if self.nvidia_visible_devices == "all":
                device = "cuda"
            else:
                device_num = self.nvidia_visible_devices.split(",")[0]
                device = f"cuda:{device_num}"

        augmented_texts = []
        for i in range(times):
            language = self.target_languages[i]
            cmd = (
                "python3 backtranslate_text.py"
                f" {context.container_input_dir / MarianMT._INPUT_FILE}"
                f" {context.container_output_dir / MarianMT._OUTPUT_FILE}"
                f" --batch-size {self.batch_size}"
                f" --cache-dir {MarianMT._CONTAINER_CACHE_DIR}"
                f" --device {device}"
                f" --marian-model {MarianMT.marian_model(language)}"
                f" --marian-inverse-model {MarianMT.marian_inverse_model(language)}"
            )

            run_kwargs = self._base_docker_run_kwargs(context)

            maybe_mount(
                run_kwargs["volumes"],
                self.host_cache_dir,
                MarianMT._CONTAINER_CACHE_DIR,
            )

            run_container(
                self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
            )

            augmented_texts.extend(self._read_output(context))

        return augmented_texts



import argparse
import ast
import itertools
import json
from pathlib import Path

import numpy as np
import pandas as pd

import torch
import torch.nn.functional as F
import transformers
from transformers import AdamW, get_linear_schedule_with_warmup


def read_unique_labels(labels_path):
    """
    Read the list of unique labels from the given path.
    """
    return labels_path.read_text(encoding="utf-8").split("\n")


def read_data(file_path, labels, multilabel):
    """
    Read some text data with optional labels from the given path.
    Return a 2-tuple of texts and labels (if any).
    """
    df = pd.read_csv(file_path, sep="\t", dtype="str", keep_default_na=False)
    X = df["Text"].tolist()
    y = None
    if labels is not None:
        if multilabel:
            y = (
                df["Label"]
                .apply(
                    lambda str_labels: [
                        1 if l in set(ast.literal_eval(str_labels)) else 0
                        for l in labels
                    ]
                )
                .tolist()
            )
        else:
            y = df["Label"].tolist()
    return X, y


def batch_gen(l, batch_size):
    for i in range(0, len(l), batch_size):
        yield l[i : i + batch_size]


def encode_batch(batch, tokenizer, max_seq_length, return_tokens=False):
    """
    Tokenize and encode the given input; stack it in a batch with fixed length
    according to the longest sequence in the batch.

    Return a tuple.  If asked to return tokens, return the encoded batch and a nested
    list of the tokenized texts.  Otherwise, just return the batch.
    """
    # Truncate input based on the sequence length
    # before stacking as a batch
    # Also return a boolean array indicating which tokens are masked
    encoded_texts = []

    for text in batch:
        encoded_text = torch.tensor(tokenizer.encode(text))[:max_seq_length]
        encoded_texts.append(encoded_text)

    encoded_batch = torch.nn.utils.rnn.pad_sequence(encoded_texts, batch_first=True)

    if return_tokens:
        tokens = [tokenizer.tokenize(text) for text in batch]
        return encoded_batch, tokens
    else:
        return (encoded_batch,)


def encode_labels(y, labels, multilabel):
    """
    Convert string labels to autoincrementing integers.
    """
    if multilabel:
        # Float type needed for proper loss computation
        return torch.as_tensor(y).float()
    else:
        mapping = {l: i for i, l in enumerate(labels)}
        return torch.as_tensor([mapping[l] for l in y])


def decode_labels(y_encoded, labels):
    """
    Convert autoincrementing integers back to strings.
    """
    return [labels[i] for i in y_encoded]


def tsv_to_encoded_batches(
    input_path,
    tokenizer,
    labels,
    batch_size,
    max_seq_length,
    multilabel,
    return_tokens=False,
):
    """
    Convert data in a csv file with optional labels to a list
    of encoded batches.  Use a generator to guard against memory errors.

    If `return_tokens` is True, add a nested list containing the tokenized
    text for each row.

    Just return encoded texts and optional tokens if no labels,
    else also return encoded labels.
    """
    has_labels = labels is not None
    X, y = read_data(input_path, labels, multilabel)

    if y is None:
        y_batch_gen = itertools.repeat(None)
    else:
        y_batch_gen = batch_gen(y, batch_size)

    for X_batch, y_batch in zip(batch_gen(X, batch_size), y_batch_gen):
        X_encoded_batch = encode_batch(
            X_batch, tokenizer, max_seq_length, return_tokens=return_tokens
        )

        if has_labels:
            y_encoded_batch = encode_labels(y_batch, labels, multilabel)
            yield (*X_encoded_batch, y_encoded_batch)
        else:
            yield X_encoded_batch


def num_batches(data_path, batch_size):
    """
    Count the number of batches contained in the given TSV data file assuming
    the given batch size.
    """
    X, _ = read_data(data_path, False, False)
    return (len(X) - 1) // batch_size + 1


def get_loss_preds(model, X, y, num_labels, multilabel):
    """
    Return the loss and predicted class(es) for each observation in the passed
    dataset.  The multilabel boolean determines how loss will be calculated and
    whether multiple predicted classes can be returned for each input observation.
    """
    if multilabel:
        outputs = model(input_ids=X)
        logits = outputs[0].view(-1, num_labels)
        loss = F.binary_cross_entropy_with_logits(logits, y)
        preds = F.sigmoid(logits.detach())
    else:
        outputs = model(input_ids=X, labels=y)
        loss = outputs[0]
        logits = outputs[1]
        # Keep preds on the device rather than moving to CPU, since we'll compare to y,
        # which is on the device
        preds = torch.argmax(logits.detach(), 1)

    return loss, preds


def train(
    *,
    input_dir,
    output_dir,
    model,
    config,
    tokenizer,
    train_batch_size,
    valid_batch_size,
    num_train_epochs,
    device,
    n_gpu,
    max_seq_length,
    labels,
    lr,
    adam_eps,
    gradient_accumulation_steps,
    multilabel,
):
    """
    Train the passed model on the given data.  Return training/validation metrics
    and save the model to the specified output directory.
    """
    # Prepare optimizer and scheduler
    # Adapted from https://github.com/huggingface/transformers/blob/master/examples/run_squad.py
    optimizer = AdamW(model.parameters(), lr=lr, eps=adam_eps)
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=0,
        num_training_steps=num_train_epochs
        * num_batches(input_dir / "train.tsv", train_batch_size)
        // gradient_accumulation_steps,
    )

    model.zero_grad()
    num_labels = len(labels)

    for epoch in range(num_train_epochs):
        print(f"Start epoch {epoch}")
        model.train()
        train_loss = 0.0
        train_count = 0

        for i, (X, y) in enumerate(
            tsv_to_encoded_batches(
                input_dir / "train.tsv",
                tokenizer,
                labels,
                train_batch_size,
                max_seq_length,
                multilabel,
            )
        ):
            X = X.to(device)
            y = y.to(device)

            loss, _ = get_loss_preds(model, X, y, num_labels, multilabel)

            if gradient_accumulation_steps > 1:
                loss /= gradient_accumulation_steps

            if n_gpu > 1:
                loss = loss.mean()  # average loss on parallel training

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)

            train_loss += loss.detach().item()
            train_count += X.size(0)

            if (i + 1) % gradient_accumulation_steps == 0:
                optimizer.step()
                scheduler.step()
                model.zero_grad()

        print(f"Epoch {epoch} train loss: {train_loss / train_count}")

        model.eval()
        valid_loss = 0.0
        valid_count = 0
        valid_correct = 0

        for X, y in tsv_to_encoded_batches(
            input_dir / "dev.tsv",
            tokenizer,
            labels,
            valid_batch_size,
            max_seq_length,
            multilabel,
        ):
            X = X.to(device)
            y = y.to(device)

            with torch.no_grad():
                loss, preds = get_loss_preds(model, X, y, num_labels, multilabel)

            if n_gpu > 1:
                loss = loss.mean()  # average loss on parallel training

            valid_loss += loss.item()
            valid_count += X.size(0)
            valid_correct += (y == preds).sum().item()

        print(f"Epoch {epoch} validation loss: {valid_loss / valid_count}")

    checkpoint_dir = output_dir / "checkpoint"
    checkpoint_dir.mkdir(exist_ok=True, parents=True)

    if hasattr(model, "module"):
        # DataParallel object -- unpack the module before saving
        model.module.save_pretrained(checkpoint_dir)
    else:
        # Plain transformers model
        model.save_pretrained(checkpoint_dir)

    tokenizer.save_pretrained(checkpoint_dir)
    config.save_pretrained(checkpoint_dir)

    if multilabel:
        valid_accuracy = valid_correct / (valid_count * num_labels)
    else:
        valid_accuracy = valid_correct / valid_count

    return {
        "mean_train_loss": train_loss / train_count,
        "mean_valid_loss": valid_loss / valid_count,
        "valid_accuracy": valid_accuracy,
    }


def predict(
    *,
    input_dir,
    output_dir,
    model,
    tokenizer,
    predict_batch_size,
    device,
    max_seq_length,
    labels,
    multilabel,
):

    pred_probas = []

    for (X,) in tsv_to_encoded_batches(
        input_dir / "test.tsv",
        tokenizer,
        None,
        predict_batch_size,
        max_seq_length,
        multilabel,
    ):
        X = X.to(device)

        with torch.no_grad():
            outputs = model(input_ids=X)
            logits = outputs[0].detach().cpu()
            if multilabel:
                pred_proba = F.sigmoid(logits)
            else:
                pred_proba = torch.softmax(logits, 1)
            pred_probas.append(pred_proba.numpy())

    df = pd.DataFrame(np.vstack(pred_probas))
    df.columns = labels
    df.to_csv(output_dir / "test_results.tsv", index=False, sep="\t")


def embed(
    *,
    input_dir,
    output_dir,
    model,
    tokenizer,
    embed_batch_size,
    device,
    max_seq_length,
    embed_pooling,
    embed_layer,
):
    embeddings = []

    with open(output_dir / "embeddings.jsonl", "w") as f:
        for X, tokens in tsv_to_encoded_batches(
            input_dir / "input.tsv",
            tokenizer,
            None,
            embed_batch_size,
            max_seq_length,
            False,
            return_tokens=True,
        ):
            X = X.to(device)

            with torch.no_grad():
                outputs = model(input_ids=X)
                hidden_states = outputs[-1]
                hidden_state = hidden_states[embed_layer]

                if embed_pooling == "mean":
                    embeddings = hidden_state.mean(dim=1).cpu().numpy().tolist()
                    for row in embeddings:
                        row_json = {"embedding": row}
                        f.write(f"{json.dumps(row_json)}\n")
                elif embed_pooling == "none":
                    embeddings = hidden_state.cpu().numpy().tolist()
                    for row, tokens in zip(embeddings, tokens):
                        # Truncate padding embeddings if needed
                        row_json = {"embedding": row[: len(tokens)], "tokens": tokens}
                        f.write(f"{json.dumps(row_json)}\n")
                else:
                    raise ValueError(f"Unsupported pooling type: {embed_pooling}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "mode",
        choices=["train", "predict", "embed"],
        help="Action to perform with the model. Train: read in a training "
        "and validation dataset, train on training and evaluate the model "
        "on validation dataset. Predict: output predictions on a test dataset. "
        "Embed: return embeddings for a test dataset.",
    )
    parser.add_argument(
        "--input-dir",
        required=True,
        help="Directory containing input files.  The exact files needed vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--output-dir",
        required=True,
        help="Directory containing output files.  The exact files created vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--cache-dir",
        required=True,
        help="Directory to use for caching transformers downloads.",
    )
    parser.add_argument(
        "--config-overrides",
        help="Path to a JSON file containing config overrides for the given model.",
    )
    parser.add_argument(
        "--model",
        required=True,
        help="Name of the transformers model to use.  If `mode` is 'train' or 'predict', "
        "this value should ensure `from transformers import <model>ModelForSequenceClassification` "
        "is a valid import. For example, model = 'Bert' -> from transformers import "
        "BertForSequenceClassification.  If `mode` is 'embed', the import will be `<model>Model`.",
    )
    parser.add_argument(
        "--weights",
        required=True,
        help="Name of the pretrained weights to use or path to a file containing "
        "pretrained weights.  If a name, must correspond to a valid weights name "
        "for the chosen model.",
    )
    parser.add_argument(
        "--multilabel",
        action="store_true",
        help="If True, model will train in a multilabel context (i.e. it's allowed to make multiple "
        "class predictions for each observation).",
    )
    parser.add_argument(
        "--train-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for training.",
    )
    parser.add_argument(
        "--valid-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for validation.",
    )
    parser.add_argument(
        "--predict-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for prediction.",
    )
    parser.add_argument(
        "--embed-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for embedding.",
    )
    parser.add_argument(
        "--num-train-epochs",
        type=int,
        default=3,
        help="Number of epochs to run training on.",
    )
    parser.add_argument(
        "--max-seq-length",
        type=int,
        default=128,
        help="Truncate sequences to this length after tokenization.  If set too large for "
        "the model used, this may cause errors.",
    )
    parser.add_argument(
        "--lr", type=float, default=5e-5, help="Learning rate for the AdamW optimizer."
    )
    parser.add_argument(
        "--adam-eps",
        type=float,
        default=1e-8,
        help="Epsilon value for the AdamW optimizer.",
    )
    parser.add_argument(
        "--gradient-accumulation-steps",
        type=int,
        default=1,
        help="Number of training steps to accumulate gradients before updating model weights.",
    )
    parser.add_argument(
        "--embed-pooling",
        choices=["mean", "none"],
        default="mean",
        help="Pooling strategy to use for combining embeddings.  If None, embeddings for "
        "each token will be returned along with the token.  Ignored if method != 'embed'.",
    )
    parser.add_argument(
        "--embed-layer",
        type=int,
        default=-2,
        help="Index of the model layer to use for generating embeddings.",
    )

    args = parser.parse_args()

    if args.gradient_accumulation_steps <= 0:
        raise ValueError(
            f"Gradient accumulation steps must be >0, got '{args.gradient_accumulation_steps}'"
        )

    input_dir = Path(args.input_dir)
    output_dir = Path(args.output_dir)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    n_gpu = torch.cuda.device_count()

    print(f"Using device: {device}")
    print(f"Number of GPUs: {n_gpu}")

    # Which model class we import depends on the task
    if args.mode in ("train", "predict"):
        model_name = f"{args.model}ForSequenceClassification"
    elif args.mode == "embed":
        model_name = f"{args.model}Model"

    tokenizer_name = f"{args.model}Tokenizer"
    config_name = f"{args.model}Config"

    print("Initializing transformer...")
    print(f"  Model: {model_name}")
    print(f"    Weights: {args.weights}")
    print(f"  Tokenizer: {tokenizer_name}")
    print(f"  Config: {config_name}")

    model_cls = getattr(transformers, model_name)
    tokenizer_cls = getattr(transformers, tokenizer_name)
    config_cls = getattr(transformers, config_name)

    tokenizer = tokenizer_cls.from_pretrained(args.weights, cache_dir=args.cache_dir)
    if tokenizer is None:
        raise ValueError("Failed to acquire tokenizer")
    config = config_cls.from_pretrained(args.weights, cache_dir=args.cache_dir)
    if config is None:
        raise ValueError("Failed to acquire config")

    # Apply config overrides specified by the user
    if args.config_overrides is not None:
        with open(args.config_overrides, "r") as f:
            config_overrides = json.load(f)
            for key, val in config_overrides.items():
                print(
                    f"Overriding config key '{key}' with value '{val}' "
                    f"(old value: '{getattr(config, key)}"
                )
                setattr(config, key, val)

    # Set the number of labels on the config object appropriately before we
    # initialize the model using the config
    # We need a list of labels for training or prediction so we know what the
    # output shape of the model should be
    labels = None
    if args.mode in ("train", "predict"):
        labels_file = input_dir / "labels.tsv"
        labels = read_unique_labels(input_dir / "labels.tsv")
        config.num_labels = len(labels)
        if config.num_labels == 1:
            raise ValueError(
                "transformers calculates regression loss when only one "
                "label is given, so it doesn't support classification when only "
                "a single label is available."
            )

        print(f"Inferred number of labels: {config.num_labels}")

    # We need to set the config to output hidden states if we're generating embeddings
    # DON'T output attentions so we can rely on hidden states being the last value in
    # the model output
    if args.mode == "embed":
        config.output_hidden_states = True
        config.output_attentions = False
        print("Outputting hidden states for embeddings")

    model = model_cls.from_pretrained(
        args.weights, cache_dir=args.cache_dir, config=config
    )
    if model is None:
        raise ValueError("Failed to acquire model")

    # If we have multiple GPUs, size the batches appropriately so DataParallel
    # distributes them correctly
    batch_multiplier = max(1, n_gpu)
    train_batch_size = args.train_batch_size * batch_multiplier
    valid_batch_size = args.valid_batch_size * batch_multiplier
    predict_batch_size = args.predict_batch_size * batch_multiplier

    model.to(device)
    if n_gpu > 1:
        model = torch.nn.DataParallel(model)

    if args.mode == "train":
        metrics = train(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            config=config,
            tokenizer=tokenizer,
            train_batch_size=train_batch_size,
            valid_batch_size=valid_batch_size,
            num_train_epochs=args.num_train_epochs,
            device=device,
            n_gpu=n_gpu,
            max_seq_length=args.max_seq_length,
            labels=labels,
            lr=args.lr,
            adam_eps=args.adam_eps,
            gradient_accumulation_steps=args.gradient_accumulation_steps,
            multilabel=args.multilabel,
        )
        with open(output_dir / "valid_results.json", "w") as f:
            json.dump(metrics, f)

    elif args.mode == "predict":
        predict(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            tokenizer=tokenizer,
            predict_batch_size=args.predict_batch_size,
            device=device,
            max_seq_length=args.max_seq_length,
            labels=labels,
            multilabel=args.multilabel,
        )
    elif args.mode == "embed":
        embed(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            tokenizer=tokenizer,
            embed_batch_size=args.embed_batch_size,
            device=device,
            max_seq_length=args.max_seq_length,
            embed_pooling=args.embed_pooling,
            embed_layer=args.embed_layer,
        )
    else:
        raise ValueError(f"invalid mode: {args.mode}")



import pytest

from gobbli.model.fasttext import FastText


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (word_ngrams)
        ({"word_ngrams": 1.0}, TypeError),
        # Bad type (lr)
        ({"lr": 1}, TypeError),
        # Bad type (dim)
        ({"dim": 100.0}, TypeError),
        # Bad type (ws)
        ({"ws": 3.0}, TypeError),
        # Bad value (fasttext_model)
        ({"fasttext_model": "bert"}, ValueError),
        # OK value (fasttext_model)
        ({"fasttext_model": "crawl-300d"}, None),
        # Dim mismatch (pretrained vectors vs user-passed dim)
        ({"fasttext_model": "crawl-300d", "dim": 100}, ValueError),
        # OK values (all)
        (
            {
                "word_ngrams": 2,
                "lr": 0.01,
                "dim": 300,
                "ws": 3,
                "fasttext_model": "crawl-300d",
            },
            None,
        ),
    ],
)
def test_init(params, exception):
    if exception is None:
        FastText(**params)
    else:
        with pytest.raises(exception):
            FastText(**params)



import argparse
import json

import tensorflow_hub as hub


def read_texts(input_file):
    with open(input_file, "r", encoding="utf-8") as f:
        return f.readlines()


def make_batches(l, batch_size):
    for i in range(0, len(l), batch_size):
        yield l[i : i + batch_size]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--input-file",
        required=True,
        help="Path to the file containing input texts, one per line.",
    )
    parser.add_argument(
        "--output-file",
        required=True,
        help="Path to write computed embeddings to (JSON format).",
    )
    parser.add_argument(
        "--module-dir",
        required=True,
        help="Path to the downloaded/extracted TFHub Module for USE.",
    )
    parser.add_argument(
        "--batch-size",
        default=32,
        type=int,
        help="Number of texts to embed at once. Default: %(default)s",
    )

    args = parser.parse_args()

    embed = hub.load(args.module_dir)
    texts = read_texts(args.input_file)

    with open(args.output_file, "w") as f:
        for batch in make_batches(texts, args.batch_size):
            embeddings = embed(batch).numpy()
            for embedding in embeddings.tolist():
                json.dump(embedding, f)
                f.write("\n")



import pytest
from spacy.lang.en import English

from gobbli.augment.wordnet import WordNet, _detokenize_doc


@pytest.mark.parametrize(
    "text",
    [
        "This is a test.",
        "Test  with double space.",
        "Test-with hyphen.",
        "Testing some 1 2 3 numbers.",
    ],
)
def test_detokenize_doc(text):
    # Initialize the spaCy extension needed to detokenize text
    WordNet()

    nlp = English()
    doc = nlp(text)

    # Fill out the replacement attribute as WordNet would.
    for tok in doc:
        tok._.replacement = tok.text
    assert _detokenize_doc(doc) == text


def test_wordnet_augment():
    wn = WordNet()
    times = 5
    new_texts = wn.augment(["This is a test."], times=times)
    assert len(new_texts) == times



import pytest

from gobbli.test.util import MockDataset


def test_base_dataset_load():
    ds = MockDataset()

    # Dataset should be unbuilt after default initialization
    assert ds._build_count == 0

    ds = MockDataset.load()

    # Dataset should now be built
    assert ds._build_count == 1

    ds.load()

    # Dataset shouldn't have been built again
    assert ds._build_count == 1


def test_base_dataset_train_input():
    # Need to build first
    with pytest.raises(ValueError):
        MockDataset().train_input()

    ds = MockDataset.load()

    # No limit
    train_input = ds.train_input(valid_proportion=0.5)

    X_len = len(MockDataset.X_TRAIN_VALID)

    assert len(train_input.X_train) == X_len / 2
    assert len(train_input.y_train) == X_len / 2
    assert len(train_input.X_valid) == X_len / 2
    assert len(train_input.y_valid) == X_len / 2

    # Limit
    train_input = ds.train_input(valid_proportion=0.5, limit=2)

    assert len(train_input.X_train) == 1
    assert len(train_input.y_train) == 1
    assert len(train_input.X_valid) == 1
    assert len(train_input.y_valid) == 1


def test_base_dataset_predict_input():
    # Need to build first
    with pytest.raises(ValueError):
        MockDataset().train_input()

    ds = MockDataset.load()

    # No limit
    predict_input = ds.predict_input()

    X_len = len(MockDataset.X_TEST)

    assert len(predict_input.X) == X_len
    assert set(predict_input.labels) == set(MockDataset.Y_TEST)

    # Limit applied
    predict_input = ds.predict_input(limit=1)

    assert len(predict_input.X) == 1

    # Make sure we only have the labels from the limited subset
    assert set(predict_input.labels) < set(MockDataset.Y_TEST)



# Copyright (c) Microsoft. All rights reserved.
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import accuracy_score, f1_score
from scipy.stats import pearsonr, spearmanr
from torch.nn.functional import cross_entropy

def compute_acc(predicts, labels):
    return 100.0 * accuracy_score(labels, predicts)

def compute_f1(predicts, labels):
    return 100.0 * f1_score(labels, predicts)

def compute_mcc(predicts, labels):
    return 100.0 * matthews_corrcoef(labels, predicts)

def compute_pearson(predicts, labels):
    pcof = pearsonr(labels, predicts)[0]
    return 100.0 * pcof

def compute_spearman(predicts, labels):
    scof = spearmanr(labels, predicts)[0]
    return 100.0 * scof

def compute_cross_entropy(predicts, labels):
    return cross_entropy(predicts, labels)



from pathlib import Path
from typing import Set, Tuple

from gobbli.dataset.nested_file import NestedFileDataset
from gobbli.util import download_archive


class IMDBDataset(NestedFileDataset):
    """
    gobbli Dataset for the IMDB sentiment analysis problem.

    https://ai.stanford.edu/~amaas/data/sentiment/
    """

    def labels(self) -> Set[str]:
        return {"pos", "neg"}

    def download(self, data_dir: Path) -> Path:
        return download_archive(
            "https://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz", data_dir
        )

    def folders(self) -> Tuple[Path, Path]:
        return Path("aclImdb/train"), Path("aclImdb/test")

    def read_source_file(self, file_path: Path) -> str:
        return file_path.read_text()



# Copyright (c) Microsoft. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

class DropoutWrapper(nn.Module):
    """
    This is a dropout wrapper which supports the fix mask dropout
    """
    def __init__(self, dropout_p=0, enable_vbp=True):
        super(DropoutWrapper, self).__init__()
        """variational dropout means fix dropout mask
        ref: https://discuss.pytorch.org/t/dropout-for-rnns/633/11
        """
        self.enable_variational_dropout = enable_vbp
        self.dropout_p = dropout_p

    def forward(self, x):
        """
            :param x: batch * len * input_size
        """
        if self.training == False or self.dropout_p == 0:
            return x

        if len(x.size()) == 3:
            mask = Variable(1.0 / (1-self.dropout_p) * torch.bernoulli((1-self.dropout_p) * (x.data.new(x.size(0), x.size(2)).zero_() + 1)), requires_grad=False)
            return mask.unsqueeze(1).expand_as(x) * x
        else:
            return F.dropout(x, p=self.dropout_p, training=self.training)



# Copyright (c) Microsoft. All rights reserved.
from copy import deepcopy
import torch
from torch.nn import Parameter
from functools import wraps

class EMA:
    def __init__(self, gamma, model):
        super(EMA, self).__init__()
        self.gamma = gamma
        self.shadow = {}
        self.model = model
        self.setup()

    def setup(self):
        for name, para in self.model.named_parameters():
            if para.requires_grad:
                self.shadow[name] = para.clone()
    def cuda(self):
        for k, v in self.shadow.items():
            self.shadow[k] = v.cuda()

    def update(self):
        for name,para in self.model.named_parameters():
            if para.requires_grad:
                self.shadow[name] = (1.0 - self.gamma) * para + self.gamma * self.shadow[name]

    def swap_parameters(self):
        for name, para in self.model.named_parameters():
            if para.requires_grad:
                temp_data = para.data
                para.data = self.shadow[name].data
                self.shadow[name].data = temp_data

    def state_dict(self):
        return self.shadow


# Adapted from
# https://github.com/pytorch/pytorch/blob/master/torch/nn/utils/weight_norm.py
# and https://github.com/salesforce/awd-lstm-lm/blob/master/weight_drop.py
def _norm(p, dim):
    """Computes the norm over all dimensions except dim"""
    if dim is None:
        return p.norm()
    elif dim == 0:
        output_size = (p.size(0),) + (1,) * (p.dim() - 1)
        return p.contiguous().view(p.size(0), -1).norm(dim=1).view(*output_size)
    elif dim == p.dim() - 1:
        output_size = (1,) * (p.dim() - 1) + (p.size(-1),)
        return p.contiguous().view(-1, p.size(-1)).norm(dim=0).view(*output_size)
    else:
        return _norm(p.transpose(0, dim), 0).transpose(0, dim)


def _dummy(*args, **kwargs):
    # We need to replace flatten_parameters with a nothing function
    return


class WeightNorm(torch.nn.Module):

    def __init__(self, weights, dim):
        super(WeightNorm, self).__init__()
        self.weights = weights
        self.dim = dim

    def compute_weight(self, module, name):
        g = getattr(module, name + '_g')
        v = getattr(module, name + '_v')
        return v * (g / _norm(v, self.dim))

    @staticmethod
    def apply(module, weights, dim):
        # Terrible temporary solution to an issue regarding compacting weights
        # re: CUDNN RNN
        if issubclass(type(module), torch.nn.RNNBase):
            module.flatten_parameters = _dummy
        if weights is None:  # do for all weight params
            weights = [w for w in module._parameters.keys() if 'weight' in w]
        fn = WeightNorm(weights, dim)
        for name in weights:
            if hasattr(module, name):
                print('Applying weight norm to {} - {}'.format(str(module), name))
                weight = getattr(module, name)
                del module._parameters[name]
                module.register_parameter(
                    name + '_g', Parameter(_norm(weight, dim).data))
                module.register_parameter(name + '_v', Parameter(weight.data))
                setattr(module, name, fn.compute_weight(module, name))

        module.register_forward_pre_hook(fn)

        return fn

    def remove(self, module):
        for name in self.weights:
            weight = self.compute_weight(module)
            delattr(module, name)
            del module._parameters[name + '_g']
            del module._parameters[name + '_v']
            module.register_parameter(name, Parameter(weight.data))

    def __call__(self, module, inputs):
        for name in self.weights:
            setattr(module, name, self.compute_weight(module, name))


def weight_norm(module, weights=None, dim=0):
    WeightNorm.apply(module, weights, dim)
    return module



from gobbli.dataset.cmu_movie_summary import MovieSummaryDataset


def test_load_cmu_movie_summary(tmp_gobbli_dir):
    ds = MovieSummaryDataset.load()

    X_train = ds.X_train()
    X_test = ds.X_test()

    y_train = ds.y_train()
    y_test = ds.y_test()

    assert len(X_train) == 33763
    assert len(y_train) == 33763
    assert len(X_test) == 8441
    assert len(y_test) == 8441

    # Ensure these objects pass validation
    train_input = ds.train_input()
    ds.predict_input()

    assert len(train_input.labels()) == 357



# coding=utf-8
# Copyright (c) Microsoft. All rights reserved.
import logging

import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable
from torch.optim.lr_scheduler import *
from data_utils.utils import AverageMeter
from pytorch_pretrained_bert import BertAdam as Adam
from module.bert_optim import Adamax
from module.my_optim import EMA
from .matcher import SANBertNetwork

logger = logging.getLogger(__name__)

class MTDNNModel(object):
    def __init__(self, opt, state_dict=None, num_train_step=-1):
        self.config = opt
        self.updates = state_dict['updates'] if state_dict and 'updates' in state_dict else 0
        self.train_loss = AverageMeter()
        self.network = SANBertNetwork(opt)

        if state_dict:
            new_state = set(self.network.state_dict().keys())
            for k in list(state_dict['state'].keys()):
                if k not in new_state:
                    del state_dict['state'][k]
            for k, v in list(self.network.state_dict().items()):
                if k not in state_dict['state']:
                    state_dict['state'][k] = v
            self.network.load_state_dict(state_dict['state'])
        self.mnetwork = nn.DataParallel(self.network) if opt['multi_gpu_on'] else self.network
        self.total_param = sum([p.nelement() for p in self.network.parameters() if p.requires_grad])

        no_decay = ['bias', 'gamma', 'beta', 'LayerNorm.bias', 'LayerNorm.weight']

        optimizer_parameters = [
            {'params': [p for n, p in self.network.named_parameters() if not any(nd in n for nd in no_decay)], 'weight_decay': 0.01},
            {'params': [p for n, p in self.network.named_parameters() if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
        ]
        
        # note that adamax are modified based on the BERT code
        if opt['optimizer'] == 'sgd':
            self.optimizer = optim.sgd(optimizer_parameters, opt['learning_rate'],
                                       weight_decay=opt['weight_decay'])

        elif opt['optimizer'] == 'adamax':
            self.optimizer = Adamax(optimizer_parameters,
                                        opt['learning_rate'],
                                        warmup=opt['warmup'],
                                        t_total=num_train_step,
                                        max_grad_norm=opt['grad_clipping'],
                                        schedule=opt['warmup_schedule'])
            if opt.get('have_lr_scheduler', False): opt['have_lr_scheduler'] = False
        elif opt['optimizer'] == 'adadelta':
            self.optimizer = optim.Adadelta(optimizer_parameters,
                                            opt['learning_rate'],
                                            rho=0.95)
        elif opt['optimizer'] == 'adam':
            self.optimizer = Adam(optimizer_parameters,
                                        lr=opt['learning_rate'],
                                        warmup=opt['warmup'],
                                        t_total=num_train_step,
                                        max_grad_norm=opt['grad_clipping'],
                                        schedule=opt['warmup_schedule'])
            if opt.get('have_lr_scheduler', False): opt['have_lr_scheduler'] = False
        else:
            raise RuntimeError('Unsupported optimizer: %s' % opt['optimizer'])

        if state_dict and 'optimizer' in state_dict:
            self.optimizer.load_state_dict(state_dict['optimizer'])

        if opt.get('have_lr_scheduler', False):
            if opt.get('scheduler_type', 'rop') == 'rop':
                self.scheduler = ReduceLROnPlateau(self.optimizer, mode='max', factor=opt['lr_gamma'], patience=3)
            elif opt.get('scheduler_type', 'rop') == 'exp':
                self.scheduler = ExponentialLR(self.optimizer, gamma=opt.get('lr_gamma', 0.95))
            else:
                milestones = [int(step) for step in opt.get('multi_step_lr', '10,20,30').split(',')]
                self.scheduler = MultiStepLR(self.optimizer, milestones=milestones, gamma=opt.get('lr_gamma'))
        else:
            self.scheduler = None
        self.ema = None
        if opt['ema_opt'] > 0:
            self.ema = EMA(self.config['ema_gamma'], self.network)
        self.para_swapped=False

    def setup_ema(self):
        if self.config['ema_opt']:
            self.ema.setup()

    def update_ema(self):
        if self.config['ema_opt']:
            self.ema.update()

    def eval(self):
        if self.config['ema_opt']:
            self.ema.swap_parameters()
            self.para_swapped = True

    def train(self):
        if self.para_swapped:
            self.ema.swap_parameters()
            self.para_swapped = False

    def update(self, batch_meta, batch_data):
        self.network.train()
        labels = batch_data[batch_meta['label']]
        if batch_meta['pairwise']:
            labels = labels.contiguous().view(-1, batch_meta['pairwise_size'])[:, 0]
        if self.config['cuda']:
            y = Variable(labels.cuda(async=True), requires_grad=False)
        else:
            y = Variable(labels, requires_grad=False)
        task_id = batch_meta['task_id']
        task_type = batch_meta['task_type']
        inputs = batch_data[:batch_meta['input_len']]
        if len(inputs) == 3:
            inputs.append(None)
            inputs.append(None)
        inputs.append(task_id)
        logits = self.mnetwork(*inputs)
        if batch_meta['pairwise']:
            logits = logits.view(-1, batch_meta['pairwise_size'])

        if self.config.get('weighted_on', False):
            if self.config['cuda']:
                weight = Variable(batch_data[batch_meta['factor']].cuda(async=True))
            else:
                weight = Variable(batch_data[batch_meta['factor']])
            if task_type > 0:
                loss = torch.mean(F.mse_loss(logits.squeeze(), y, reduce=False) * weight)
            else:
                loss = torch.mean(F.cross_entropy(logits, y, reduce=False) * weight)
        else:
            if task_type > 0:
                loss = F.mse_loss(logits.squeeze(), y)
            else:
                loss = F.cross_entropy(logits, y)

        self.train_loss.update(loss.item(), logits.size(0))
        self.optimizer.zero_grad()

        loss.backward()
        if self.config['global_grad_clipping'] > 0:
            torch.nn.utils.clip_grad_norm_(self.network.parameters(),
                                          self.config['global_grad_clipping'])
        self.optimizer.step()
        self.updates += 1
        self.update_ema()

    def predict(self, batch_meta, batch_data):
        self.network.eval()
        task_id = batch_meta['task_id']
        task_type = batch_meta['task_type']
        inputs = batch_data[:batch_meta['input_len']]
        if len(inputs) == 3:
            inputs.append(None)
            inputs.append(None)
        inputs.append(task_id)
        score = self.mnetwork(*inputs)
        if batch_meta['pairwise']:
            score = score.contiguous().view(-1, batch_meta['pairwise_size'])
            if task_type < 1:
                score = F.softmax(score, dim=1)
            score = score.data.cpu()
            score = score.numpy()
            predict = np.zeros(score.shape, dtype=int)
            positive = np.argmax(score, axis=1)
            for idx, pos in enumerate(positive):
                predict[idx, pos] = 1
            predict = predict.reshape(-1).tolist()
            score = score.reshape(-1).tolist()
            return score, predict, batch_meta['true_label']
        else:
            if task_type < 1:
                score = F.softmax(score, dim=1)
            score = score.data.cpu()
            score = score.numpy()
            predict = np.argmax(score, axis=1).tolist()
            score = score.reshape(-1).tolist()
        return score, predict, batch_meta['label']

    def save(self, filename):
        network_state = dict([(k, v.cpu()) for k, v in self.network.state_dict().items()])
        ema_state = dict(
            [(k, v.cpu()) for k, v in self.ema.model.state_dict().items()]) if self.ema is not None else dict()
        params = {
            'state': network_state,
            'optimizer': self.optimizer.state_dict(),
            'ema': ema_state,
            'config': self.config,
        }
        torch.save(params, filename)
        logger.info('model saved to {}'.format(filename))

    def cuda(self):
        self.network.cuda()
        if self.config['ema_opt']:
            self.ema.cuda()



from pathlib import Path

import joblib
import pytest
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.svm import LinearSVC

from gobbli.model.sklearn import (
    SKLearnClassifier,
    make_cv_tfidf_logistic_regression,
    make_default_tfidf_logistic_regression,
    persist_estimator,
)
from gobbli.model.sklearn.model import _AT_LEAST_TWO_CLASSES_ERR_MSG, _SafeEstimator


def _make_test_estimator():
    # Don't use anything complicated, like a Pipeline or GridSearchCV,
    # since those may have estimator params which don't compare cleanly
    return LogisticRegression()


def _assert_estimators_equal(clf1, clf2):
    assert type(clf1) is type(clf2)
    assert clf1.get_params() == clf2.get_params()


def test_dump_estimator(tmpdir):
    tempdir_path = Path(tmpdir)
    clf = _make_test_estimator()

    dump_path = tempdir_path / "test.joblib"
    SKLearnClassifier._dump_estimator(clf, dump_path)
    loaded_clf = joblib.load(dump_path)

    _assert_estimators_equal(clf, loaded_clf)


def test_load_estimator(tmpdir):
    tempdir_path = Path(tmpdir)
    clf = _make_test_estimator()

    dump_path = tempdir_path / "test.joblib"
    joblib.dump(clf, dump_path)
    loaded_clf = SKLearnClassifier._load_estimator(dump_path)

    _assert_estimators_equal(clf, loaded_clf)


def test_persist_estimator(tmp_gobbli_dir):
    clf = _make_test_estimator()
    clf_path = persist_estimator(clf)

    assert tmp_gobbli_dir in clf_path.parents

    loaded_clf = SKLearnClassifier._load_estimator(clf_path)

    _assert_estimators_equal(clf, loaded_clf)


@pytest.mark.parametrize(
    "clf,err",
    [
        # Not an estimator
        (None, ValueError),
        # Also not an estimator
        (1, ValueError),
        # Estimator but not a classifier
        (LinearRegression(), ValueError),
        # Estimator with no predict_proba
        (LinearSVC(), ValueError),
        # Valid estimator
        (LogisticRegression(), None),
        # Invalid pipeline (not a classifier)
        (Pipeline([("linreg", LinearRegression())]), ValueError),
        # Invalid pipeline (no predict_proba)
        (Pipeline([("svc", LinearSVC())]), ValueError),
        # Valid pipeline
        (Pipeline([("logreg", LogisticRegression())]), None),
        # Invalid grid search (not a classifier)
        (GridSearchCV(LinearRegression(), {}), ValueError),
        # Invalid grid search (no predict_proba)
        (GridSearchCV(LinearSVC(), {}), ValueError),
        # Valid grid search
        (GridSearchCV(LogisticRegression(), {}), None),
        # Our helpers should both be valid
        (make_cv_tfidf_logistic_regression(), None),
        (make_default_tfidf_logistic_regression(), None),
    ],
)
def test_validate_estimator(clf, err):
    if err is not None:
        with pytest.raises(err):
            SKLearnClassifier._validate_estimator(clf)
    else:
        SKLearnClassifier._validate_estimator(clf)


def test_safe_estimator():
    clf = make_default_tfidf_logistic_regression()

    X_train = ["test", "test2"]
    y_train = ["a", "a"]

    with pytest.raises(ValueError) as e:
        clf.fit(X_train, y_train)
    assert _AT_LEAST_TWO_CLASSES_ERR_MSG in str(e.value)

    safe_clf = _SafeEstimator(clf)
    safe_clf.fit(X_train, y_train)
    assert safe_clf.classes_.tolist() == ["a"]

    y_pred = safe_clf.predict(X_train)
    assert y_pred.tolist() == ["a", "a"]

    y_pred_proba = safe_clf.predict_proba(X_train)
    assert y_pred_proba.tolist() == [[1], [1]]


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (estimator_path)
        ({"estimator_path": 1}, TypeError),
        # init loads the model path, and there isn't a good way to
        # reference a temp path from this param list, so we'll assume
        # other tests will catch a failure to initialize from a good
        # estimator path
    ],
)
def test_init(params, exception):
    with pytest.raises(exception):
        SKLearnClassifier(**params)



# coding=utf-8
# Copyright (c) Microsoft. All rights reserved.
import os
import sys
import json
import torch
import random
import string
import logging
import numpy as np
import pickle as pkl
from shutil import copyfile

UNK_ID=100
BOS_ID=101

class BatchGen:
    def __init__(self, data, batch_size=32, gpu=True, is_train=True,
                 maxlen=128, dropout_w=0.005,
                 do_batch=True, weighted_on=False,
                 task_id=0,
                 pairwise=False,
                 task=None,
                 task_type=0,
                 data_type=0):
        self.batch_size = batch_size
        self.maxlen = maxlen
        self.is_train = is_train
        self.gpu = gpu
        self.weighted_on = weighted_on
        self.data = data
        self.task_id = task_id
        self.pairwise = pairwise
        self.pairwise_size = 1
        self.data_type = data_type
        self.task_type=task_type
        if do_batch:
            if is_train:
                indices = list(range(len(self.data)))
                random.shuffle(indices)
                data = [self.data[i] for i in indices]
            self.data = BatchGen.make_baches(data, batch_size)
        self.offset = 0
        self.dropout_w = dropout_w

    @staticmethod
    def make_baches(data, batch_size=32):
        return [data[i:i + batch_size] for i in range(0, len(data), batch_size)]

    @staticmethod
    def load(path, is_train=True, maxlen=128, factor=1.0, pairwise=False):
        with open(path, 'r', encoding='utf-8') as reader:
            data = []
            cnt = 0
            for line in reader:
                sample = json.loads(line)
                sample['factor'] = factor
                cnt += 1
                if is_train:
                    if pairwise and (len(sample['token_id'][0]) > maxlen or len(sample['token_id'][1]) > maxlen):
                        continue
                    if (not pairwise) and (len(sample['token_id']) > maxlen):
                        continue
                data.append(sample)
            print('Loaded {} samples out of {}'.format(len(data), cnt))
            return data

    def reset(self):
        if self.is_train:
            indices = list(range(len(self.data)))
            random.shuffle(indices)
            self.data = [self.data[i] for i in indices]
        self.offset = 0

    def __random_select__(self, arr):
        if self.dropout_w > 0:
            return [UNK_ID if random.uniform(0, 1) < self.dropout_w else e for e in arr]
        else: return arr

    def __len__(self):
        return len(self.data)

    def patch(self, v):
        v = v.cuda(async=True)
        return v

    @staticmethod
    def todevice(v, device):
        v = v.to(device)
        return v

    def rebacth(self, batch):
        newbatch = []
        for sample in batch:
            size = len(sample['token_id'])
            self.pairwise_size = size
            assert size == len(sample['type_id'])
            for idx in range(0, size):
                token_id = sample['token_id'][idx]
                type_id = sample['type_id'][idx]
                uid = sample['ruid'][idx]
                olab = sample['olabel'][idx]
                newbatch.append({'uid': uid, 'token_id': token_id, 'type_id': type_id, 'label':sample['label'], 'true_label': olab})
        return newbatch


    def __iter__(self):
        while self.offset < len(self):
            batch = self.data[self.offset]
            if self.pairwise:
                batch = self.rebacth(batch)
            batch_size = len(batch)
            batch_dict = {}
            tok_len = max(len(x['token_id']) for x in batch)
            hypothesis_len = max(len(x['type_id']) - sum(x['type_id']) for x in batch)
            token_ids = torch.LongTensor(batch_size, tok_len).fill_(0)
            type_ids = torch.LongTensor(batch_size, tok_len).fill_(0)
            masks = torch.LongTensor(batch_size, tok_len).fill_(0)
            if self.data_type < 1:
                premise_masks = torch.ByteTensor(batch_size, tok_len).fill_(1)
                hypothesis_masks = torch.ByteTensor(batch_size, hypothesis_len).fill_(1)

            for i, sample in enumerate(batch):
                select_len = min(len(sample['token_id']), tok_len)
                tok = sample['token_id']
                if self.is_train:
                    tok = self.__random_select__(tok)
                token_ids[i, :select_len] = torch.LongTensor(tok[:select_len])
                type_ids[i, :select_len] = torch.LongTensor(sample['type_id'][:select_len])
                masks[i, :select_len] = torch.LongTensor([1] * select_len)
                if self.data_type < 1:
                    hlen = len(sample['type_id']) - sum(sample['type_id'])
                    hypothesis_masks[i, :hlen] = torch.LongTensor([0] * hlen)
                    for j in range(hlen, select_len):
                        premise_masks[i, j] = 0
            if self.data_type < 1:
                batch_info = {
                    'token_id': 0,
                    'segment_id': 1,
                    'mask': 2,
                    'premise_mask': 3,
                    'hypothesis_mask': 4
                    }
                batch_data = [token_ids, type_ids, masks, premise_masks, hypothesis_masks]
                current_idx = 5
                valid_input_len = 5
            else:
                batch_info = {
                    'token_id': 0,
                    'segment_id': 1,
                    'mask': 2
                    }
                batch_data = [token_ids, type_ids, masks]
                current_idx = 3
                valid_input_len = 3

            if self.is_train:
                labels = [sample['label'] for sample in batch]
                if self.task_type > 0:
                    batch_data.append(torch.FloatTensor(labels))
                else:
                    batch_data.append(torch.LongTensor(labels))
                batch_info['label'] = current_idx
                current_idx += 1

            if self.gpu:
                for i, item in enumerate(batch_data):
                    batch_data[i] = self.patch(item.pin_memory())

            # meta 
            batch_info['uids'] = [sample['uid'] for sample in batch]
            batch_info['task_id'] = self.task_id
            batch_info['input_len'] = valid_input_len
            batch_info['pairwise'] = self.pairwise
            batch_info['pairwise_size'] = self.pairwise_size
            batch_info['task_type'] = self.task_type
            if not self.is_train:
                labels = [sample['label'] for sample in batch]
                batch_info['label'] = labels
                if self.pairwise:
                    batch_info['true_label'] = [sample['true_label'] for sample in batch]
            self.offset += 1
            yield batch_info, batch_data



from gobbli.dataset.base import BaseDataset


class TrivialDataset(BaseDataset):
    """
    gobbli Dataset containing only a few observations.
    Useful for verifying a model runs without waiting for an
    actual dataset to process.
    """

    DATASET = ["This is positive.", "This, although, is negative."]
    LABELS = ["1", "0"]

    def _is_built(self) -> bool:
        return True

    def _build(self):
        pass

    def X_train(self):
        return TrivialDataset.DATASET

    def y_train(self):
        return TrivialDataset.LABELS

    def X_test(self):
        return TrivialDataset.DATASET

    def y_test(self):
        return TrivialDataset.LABELS



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from gobbli.util import gobbli_dir


def augment_dir() -> Path:
    return gobbli_dir() / "augment"


class BaseAugment(ABC):
    """
    Base class for data augmentation methods.
    """

    @abstractmethod
    def augment(self, X: List[str], times: int = 5, p: float = 0.1) -> List[str]:
        """
        Return additional texts for each text in the passed array.

        Args:
          X: Input texts.
          times: How many texts to generate per text in the input.
          p: Probability of considering each token in the input for replacement.
            Note that some tokens aren't able to be replaced by a given augmentation
            method and will be ignored, so the actual proportion of replaced tokens
            in your input may be much lower than this number.
        Returns:
          Generated texts (length = ``times * len(X)``).
        """
        raise NotImplementedError

    @classmethod
    def data_dir(cls) -> Path:
        """
        Returns:
          The data directory used for this class of augmentation model.
        """
        return augment_dir() / cls.__name__
